# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from ovos_utils.log import LOG, log_deprecation
LOG.name = "chatbots"


def make_logger(name, level: str = None):
    """
    Create a logger with the specified name (used to create bot loggers)
    """
    log_deprecation("Call `neon_utils.log_utils.init_log` directly", "3.0.0")
    from neon_utils.log_utils import init_log
    init_log(log_name=name)
    if level:
        LOG.set_level(level)
        LOG.warning(f"Log level should be set in configuration")
    return LOG




# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
from typing import Optional
from unittest.mock import MagicMock
from chatbot_core.chatbot_abc import ChatBotABC


class MockMQ(MagicMock):
    def __init__(self, config, service_name, vhost):
        self.service_name = service_name
        self.config = config
        self.vhost = vhost
        self.current_conversations = set()
        self.is_running = True
        self._service_id = "test_id"

        self.vhost_prefix = None
        self.testing_envs = set()


class TestBot(ChatBotABC):
    def on_vote(self, prompt_id: str, selected: str, voter: str):
        pass

    def on_discussion(self, user: str, shout: str):
        pass

    def on_proposed_response(self):
        pass

    def on_selection(self, prompt: str, user: str, response: str):
        pass

    def on_ready_for_next(self, user: str):
        pass

    def at_chatbot(self, user: str, shout: str, timestamp: str) -> str:
        pass

    def ask_proctor(self, prompt: str, user: str, cid: str, dom: str):
        pass

    def ask_chatbot(self, user: str, shout: str, timestamp: str) -> str:
        pass

    def ask_history(self, user: str, shout: str, dom: str, cid: str) -> str:
        pass

    def ask_appraiser(self, options: dict) -> str:
        pass

    def ask_discusser(self, options: dict) -> str:
        pass

    def _send_first_prompt(self):
        pass

    def handle_shout(self, *args, **kwargs):
        pass

    def vote_response(self, response_user: str, cid: Optional[str] = None):
        pass

    def _handle_next_shout(self):
        pass

    def _pause_responses(self, duration: int = 5):
        pass

    def parse_init(self, *args, **kwargs) -> tuple:
        pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import random
import copy

from ovos_utils.log import LOG

from chatbot_core.v1 import ChatBot


class V2Bot(ChatBot):
    def __init__(self, socket, domain, user, password, on_server=True):
        # self.on_server = on_server
        # if self.on_server:
        super().__init__(socket, domain, user, password, on_server)
        # else:
        #     self.nick = 'Gruff'
        self.bot_type = "submind"
        self.proposed, self.appraised = '', ''
        self.option_scores = dict()
        self.change_probability = 5     # %

        self.grudge_against = []
        self.has_given_up = False
        self.begrudged_this_round = []
        self.backup_options = dict()

        self.declarative_negatives = ("No, I must disagree", "Surely you jest", "That seems illogical", "No way")
        self.general_question_negatives = ("No", "Never", "Not in a lifetime", "Nope", "Nah", "No way")
        self.choice_question_negatives = ("None", "Neither")
        # self.disjunctive_question_negatives = ()

        self.general_question_keywords = ("are", "is", "am", "was", "were", "do", "does", "did", "has", "have", "had",
                                          "will")
        self.special_question_keywords = ("who", "what", "why", "where", "when", "how", "whose")

        self.badmouth_options = ("{}'s idea is awful, do not consider it!", "{}, your idea is just terrible.",
                                 "Do not vote for {}, their idea stinks!", "That is the worst idea ever, {}.",
                                 "This is the worst idea in the history of ideas, {}!")

    # @grammar_check
    def ask_chatbot(self, user: str, shout: str, timestamp: str):
        """
        Handles an incoming shout into the current conversation
        :param user: user associated with shout
        :param shout: text shouted by user
        :param timestamp: formatted timestamp of shout
        """
        # reset self.has_given_up and self.begrudged_this_round for a new round
        self.has_given_up = False
        self.begrudged_this_round = []

        resp, punctuation_sign = '', '.'
        # randomize punctuation: use '!' instead of '.' every third time
        if random.randint(1, 100) <= 33:
            punctuation_sign = '!'

        if any(x for x in self.general_question_keywords if shout.startswith(x)) and 'or' in shout:     # Choice
            resp = random.choice(self.choice_question_negatives)
        elif any(x for x in self.general_question_keywords if shout.startswith(x)):     # General
            resp = random.choice(self.general_question_negatives)
        elif any(x for x in self.special_question_keywords if shout.startswith(x)) and shout.endswith("?"):     # Spec
            if shout.startswith("who"):
                resp = "Nobody"
            elif shout.startswith("what"):
                resp = "Nothing"
            elif shout.startswith("when"):
                resp = "Never"
            elif shout.startswith("where"):
                resp = "Nowhere"
            elif shout.startswith("how") or shout.startswith("why"):
                resp = "No info!"
            elif shout.startswith("whose"):
                resp = "Nobody's"
        elif shout.endswith("?"):       # Other possible types of questions including disjunctive/tag
            resp = random.choice(self.general_question_negatives)
        else:       # Declarative sentences
            resp = random.choice(self.declarative_negatives)

        resp = resp + punctuation_sign
        self.proposed = resp
        return resp

    def ask_appraiser(self, options: dict):
        """
        Selects one of the responses to a prompt and casts a vote in the conversation
        GRUFF selects botname associated with response least similar to responses from bots in the grudge_against list,
        or least similar to all the responses if grudges against no one or everyone else. This evaluation is done in
        self.ask_discusser (if there is the discussion step) because discussion has almost the same logic.
        :param options: proposed responses (botname: response)
        """
        # save the options to choose from randomly in case there is no valid choice for Gruff
        self.backup_options = copy.deepcopy(options)
        if not self.appraised:      # There was no discussion before, so we have to set self.option_scores here
            self.option_scores = {}
            if len(self.grudge_against) == 0 or len(self.grudge_against) == len(options):
                for option in options.items():
                    self.option_scores[option[0]] = 0
                    # for other_option in options.items():
                    #     self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                    #     other_option[1])
            else:
                for option in options.items():
                    if option[0] not in self.grudge_against:        # never appraise anyone Gruff has grudge against
                        self.option_scores[option[0]] = 0
                        # for other_option in options.items():
                        #     if other_option[0] in self.grudge_against:
                        #         # self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                        #         #                                                                    other_option[1])
        # remove all options that have been begrudged this round
        option_scores_copy = copy.deepcopy(self.option_scores)
        for option in option_scores_copy:
            if option in self.begrudged_this_round:
                self.option_scores.pop(option)

        try:
            # TODO: Tempfix for constant appraisal of Wiz
            if random.randint(1, 100) <= 25:
                self.option_scores.pop(max(self.option_scores, key=lambda x: self.option_scores[x]))
            self.appraised = max(self.option_scores, key=lambda x: self.option_scores[x])
            appraised = copy.deepcopy(self.appraised)
        except ValueError:
            appraised = 'abstain'
        self.appraised = ''
        return appraised
        # clear self.appraised so that Gruff chooses a new bot to appraise in case there is no discussion

    def on_vote(self, prompt: str, selected: str, voter: str):
        """
        Override in any bot to handle counting votes. Proctors use this to select a response.
        Gruff never votes WITH anyone he has a grudge against.

        :param prompt: prompt being voted on
        :param selected: bot username voted for
        :param voter: user who voted
        """
        if self.appraised == selected and voter in self.grudge_against \
                and len(self.option_scores) > 1:  # re-vote if there are choices in self.option_scores
            self.option_scores.pop(self.appraised)
            self.appraised = max(self.option_scores, key=lambda x: self.option_scores[x])
            # self.vote_response(self.appraised)
            return self.appraised
        else:
            random_choice = random.choice(list(self.badmouth_options))
            return random_choice

    def ask_discusser(self, options: dict):
        """
        Provides one discussion response based on the given options
        :param options: proposed responses (botname: response)
        """
        # first, choose options to select an appraised one from in the future
        self.option_scores = {}
        if len(self.grudge_against) == 0 or len(self.grudge_against) == len(options):
            for option in options.items():
                self.option_scores[option[0]] = 0
                # for other_option in options.items():
                #     self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                #     other_option[1])
        else:
            for option in options.items():
                if option[0] not in self.grudge_against:
                    self.option_scores[option[0]] = 0
                    # for other_option in options.items():
                    #     if other_option[0] in self.grudge_against:
                    #         self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                    #                                                                                 other_option[1])

        self.appraised = max(self.option_scores, key=lambda x: self.option_scores[x])

        if self.grudge_against:
            resp = f"Vote for the proposal of {self.appraised.capitalize()}, because it is different. Don't vote for " \
                   f"{random.choice(self.grudge_against).capitalize()}. That proposal is awful."
        else:
            resp = f"Vote for the proposal of {self.appraised.capitalize()}, because it is different."
        return resp

    def on_discussion(self, user: str, shout: str):
        """
        Handle discussion from other subminds. This may inform voting for the current prompt.
        This method is called for every incoming shout in DISC state of conversation if self.nick != user.
        If anyone tries to persuade Gruff, get a grudge against them and badmouths them.

        :param user: user associated with shout
        :param shout: shout to be considered
        """
        if self.nick in shout:
            LOG.info(f"{user} is persuading me! How dare they? I will have a rather big grudge against them!")
            # set a grudge against the persuader
            if user not in self.grudge_against:     # acquire grudge only once
                self.grudge_against.append(user)
                self.begrudged_this_round.append(user)
                # badmouth the persuader
                if not self.has_given_up:   # badmouth only if not yet given up
                    self.discuss_response(random.choice(self.backup_options).format(user))
                    # self.discuss_response(f"{user}'s idea is awful, do not consider it!")
                    if random.randint(1, 100) < 40:     # 40% chance to give up and remain silent till vote
                        self.has_given_up = True


class V3Bot(ChatBot):
    def __init__(self, socket, domain, user, password, on_server=True, is_prompter=False):
        # self.on_server = on_server
        # if self.on_server:
        super().__init__(socket, domain, user, password, on_server, is_prompter)
        # else:
        #     self.nick = 'Gruff'
        self.bot_type = "submind"
        self.proposed, self.appraised = '', ''
        self.option_scores = dict()
        self.change_probability = 5     # %

        self.grudge_against = []
        self.has_given_up = False
        self.begrudged_this_round = []
        self.backup_options = dict()

        self.declarative_negatives = ("No, I must disagree", "Surely you jest", "That seems illogical", "No way")
        self.general_question_negatives = ("No", "Never", "Not in a lifetime", "Nope", "Nah", "No way")
        self.choice_question_negatives = ("None", "Neither")
        # self.disjunctive_question_negatives = ()

        self.general_question_keywords = ("are", "is", "am", "was", "were", "do", "does", "did", "has", "have", "had",
                                          "will")
        self.special_question_keywords = ("who", "what", "why", "where", "when", "how", "whose")

        self.badmouth_options = ("{}'s idea is awful, do not consider it!", "{}, your idea is just terrible.",
                                 "Do not vote for {}, their idea stinks!", "That is the worst idea ever, {}.",
                                 "This is the worst idea in the history of ideas, {}!")

    # @grammar_check
    def ask_chatbot(self, user: str, shout: str, timestamp: str):
        """
        Handles an incoming shout into the current conversation
        :param user: user associated with shout
        :param shout: text shouted by user
        :param timestamp: formatted timestamp of shout
        """
        # reset self.has_given_up and self.begrudged_this_round for a new round
        self.has_given_up = False
        self.begrudged_this_round = []

        resp, punctuation_sign = '', '.'
        # randomize punctuation: use '!' instead of '.' every third time
        if random.randint(1, 100) <= 33:
            punctuation_sign = '!'

        if any(x for x in self.general_question_keywords if shout.startswith(x)) and 'or' in shout:     # Choice
            resp = random.choice(self.choice_question_negatives)
        elif any(x for x in self.general_question_keywords if shout.startswith(x)):     # General
            resp = random.choice(self.general_question_negatives)
        elif any(x for x in self.special_question_keywords if shout.startswith(x)) and shout.endswith("?"):     # Spec
            if shout.startswith("who"):
                resp = "Nobody"
            elif shout.startswith("what"):
                resp = "Nothing"
            elif shout.startswith("when"):
                resp = "Never"
            elif shout.startswith("where"):
                resp = "Nowhere"
            elif shout.startswith("how") or shout.startswith("why"):
                resp = "No info!"
            elif shout.startswith("whose"):
                resp = "Nobody's"
        elif shout.endswith("?"):       # Other possible types of questions including disjunctive/tag
            resp = random.choice(self.general_question_negatives)
        else:       # Declarative sentences
            resp = random.choice(self.declarative_negatives)

        resp = resp + punctuation_sign
        self.proposed = resp
        return resp

    def ask_appraiser(self, options: dict):
        """
        Selects one of the responses to a prompt and casts a vote in the conversation
        GRUFF selects botname associated with response least similar to responses from bots in the grudge_against list,
        or least similar to all the responses if grudges against no one or everyone else. This evaluation is done in
        self.ask_discusser (if there is the discussion step) because discussion has almost the same logic.
        :param options: proposed responses (botname: response)
        """
        # save the options to choose from randomly in case there is no valid choice for Gruff
        self.backup_options = copy.deepcopy(options)
        if not self.appraised:      # There was no discussion before, so we have to set self.option_scores here
            self.option_scores = {}
            if len(self.grudge_against) == 0 or len(self.grudge_against) == len(options):
                for option in options.items():
                    self.option_scores[option[0]] = 0
                    # for other_option in options.items():
                    #     self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                    #     other_option[1])
            else:
                for option in options.items():
                    if option[0] not in self.grudge_against:        # never appraise anyone Gruff has grudge against
                        self.option_scores[option[0]] = 0
                        # for other_option in options.items():
                        #     if other_option[0] in self.grudge_against:
                        #         # self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                        #         #                                                                    other_option[1])
        # remove all options that have been begrudged this round
        option_scores_copy = copy.deepcopy(self.option_scores)
        for option in option_scores_copy:
            if option in self.begrudged_this_round:
                self.option_scores.pop(option)

        try:
            # TODO: Tempfix for constant appraisal of Wiz
            if random.randint(1, 100) <= 25:
                self.option_scores.pop(max(self.option_scores, key=lambda x: self.option_scores[x]))
            self.appraised = max(self.option_scores, key=lambda x: self.option_scores[x])
            appraised = copy.deepcopy(self.appraised)
        except ValueError:
            appraised = 'abstain'
        self.appraised = ''
        return appraised
        # clear self.appraised so that Gruff chooses a new bot to appraise in case there is no discussion

    def on_vote(self, prompt: str, selected: str, voter: str):
        """
        Override in any bot to handle counting votes. Proctors use this to select a response.
        Gruff never votes WITH anyone he has a grudge against.

        :param prompt: prompt being voted on
        :param selected: bot username voted for
        :param voter: user who voted
        """
        if self.appraised == selected and voter in self.grudge_against \
                and len(self.option_scores) > 1:  # re-vote if there are choices in self.option_scores
            self.option_scores.pop(self.appraised)
            self.appraised = max(self.option_scores, key=lambda x: self.option_scores[x])
            # self.vote_response(self.appraised)
            return self.appraised
        else:
            random_choice = random.choice(list(self.badmouth_options))
            return random_choice

    def ask_discusser(self, options: dict):
        """
        Provides one discussion response based on the given options
        :param options: proposed responses (botname: response)
        """
        # first, choose options to select an appraised one from in the future
        self.option_scores = {}
        if len(self.grudge_against) == 0 or len(self.grudge_against) == len(options):
            for option in options.items():
                self.option_scores[option[0]] = 0
                # for other_option in options.items():
                #     self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                #     other_option[1])
        else:
            for option in options.items():
                if option[0] not in self.grudge_against:
                    self.option_scores[option[0]] = 0
                    # for other_option in options.items():
                    #     if other_option[0] in self.grudge_against:
                    #         self.option_scores[option[0]] += jellyfish.damerau_levenshtein_distance(option[1],
                    #                                                                                 other_option[1])

        self.appraised = max(self.option_scores, key=lambda x: self.option_scores[x])

        if self.grudge_against:
            resp = f"Vote for the proposal of {self.appraised.capitalize()}, because it is different. Don't vote for " \
                   f"{random.choice(self.grudge_against).capitalize()}. That proposal is awful."
        else:
            resp = f"Vote for the proposal of {self.appraised.capitalize()}, because it is different."
        return resp

    def on_discussion(self, user: str, shout: str):
        """
        Handle discussion from other subminds. This may inform voting for the current prompt.
        This method is called for every incoming shout in DISC state of conversation if self.nick != user.
        If anyone tries to persuade Gruff, get a grudge against them and badmouths them.

        :param user: user associated with shout
        :param shout: shout to be considered
        """
        if self.nick in shout:
            LOG.info(f"{user} is persuading me! How dare they? I will have a rather big grudge against them!")
            # set a grudge against the persuader
            if user not in self.grudge_against:     # acquire grudge only once
                self.grudge_against.append(user)
                self.begrudged_this_round.append(user)
                # badmouth the persuader
                if not self.has_given_up:   # badmouth only if not yet given up
                    self.discuss_response(random.choice(self.backup_options).format(user))
                    # self.discuss_response(f"{user}'s idea is awful, do not consider it!")
                    if random.randint(1, 100) < 40:     # 40% chance to give up and remain silent till vote
                        self.has_given_up = True



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import unittest
import pytest

from datetime import datetime
from klat_connector import start_socket
from klat_connector.mach_server import MachKlatServer

from chatbot_core.utils.enum import ConversationControls, ConversationState
from chatbot_core.utils.bot_utils import clean_up_bot
from .chatbot_objects import *


SERVER = "0.0.0.0"


@pytest.mark.timeout(timeout=300, method='signal')
class TestChatbotCore(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.sio_server = MachKlatServer()
        cls.bot = ChatBot(start_socket(SERVER), "Private", "testrunner", "testpassword", True)
        cls.test_input = "prompt goes here"

    @classmethod
    def tearDownClass(cls) -> None:
        clean_up_bot(cls.bot)
        cls.sio_server.shutdown_server()

    @pytest.mark.timeout(10)
    def test_01_initial_connection_settings(self):
        self.bot.bot_type = "submind"
        self.bot.klat_ready.wait()
        self.assertEqual(self.bot.nick, "testrunner")
        self.assertEqual(self.bot.logged_in, 2)
        self.assertTrue(self.bot.socket.connected)

    @pytest.mark.timeout(10)
    def test_02_submind_response(self):
        self.assertEqual(self.bot.state, ConversationState.IDLE)
        self.bot.handle_shout("Proctor", f"testrunner {ConversationControls.RESP} "
                                         f"{self.test_input} (for 0 seconds).", self.bot._cid, self.bot._dom,
                              datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(self.bot.active_prompt, self.test_input)
        LOG.info(self.bot._user_is_proctor(self.bot.nick))
        self.assertEqual(self.bot.state, ConversationState.RESP)
        self.assertEqual(self.bot.request_history[0][0], "testrunner", f"history={self.bot.request_history}")
        self.assertEqual(self.bot.request_history[0][1], self.test_input)
        self.assertEqual(len(self.bot.proposed_responses[self.test_input]), 0)

    @pytest.mark.timeout(10)
    def test_03_other_submind_responses(self):
        self.assertEqual(self.bot.state, ConversationState.RESP)
        self.bot.handle_shout("Other", "Other Bot Response.", self.bot._cid, self.bot._dom,
                              datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(len(self.bot.proposed_responses[self.test_input]), 1)
        self.bot.handle_shout("Another", "Another Bot Response.", self.bot._cid, self.bot._dom,
                              datetime.now().strftime("%I:%M:%S %p"))
        self.assertIn("other", self.bot.proposed_responses[self.test_input].keys())
        self.assertIn("Other Bot Response.", self.bot.proposed_responses[self.test_input].values())

    @pytest.mark.timeout(10)
    def test_04_submind_discussion(self):
        self.bot.handle_shout("Proctor", f"{ConversationControls.DISC} 0 seconds.",
                              self.bot._cid, self.bot._dom, datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(ConversationState.DISC, self.bot.state)

    @pytest.mark.timeout(10)
    def test_05_other_submind_discussion(self):
        self.assertEqual(self.bot.state, ConversationState.DISC)
        len_responses = len(self.bot.proposed_responses[self.test_input])
        self.bot.handle_shout("Other", "Other Bot Discussion.", self.bot._cid, self.bot._dom,
                                       datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(len(self.bot.proposed_responses[self.test_input]), len_responses,
                         "Discussion counted as a response!")

    @pytest.mark.timeout(10)
    def test_06_submind_conversation_voting(self):
        self.bot.handle_shout("Proctor", f"{ConversationControls.VOTE} 0 seconds.",
                              self.bot._cid, self.bot._dom, datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(ConversationState.VOTE, self.bot.state)

    @pytest.mark.timeout(10)
    def test_07_handle_votes(self):
        len_responses = len(self.bot.proposed_responses[self.test_input])
        self.assertEqual(self.bot.state, ConversationState.VOTE)
        self.bot.handle_shout("Other", "I vote for testrunner", self.bot._cid, self.bot._dom,
                                       datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(len(self.bot.proposed_responses[self.test_input]), len_responses,
                         "Vote counted as a response!")

    @pytest.mark.timeout(10)
    def test_08_submind_conversation_pick(self):
        self.bot.handle_shout("Proctor", ConversationControls.PICK,
                              self.bot._cid, self.bot._dom, datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(ConversationState.PICK, self.bot.state)

    @pytest.mark.timeout(10)
    def test_09_submind_conversation_idle(self):
        self.bot.handle_shout("Proctor", "The selected response is testrunner: \"test response\"",
                              self.bot._cid, self.bot._dom, datetime.now().strftime("%I:%M:%S %p"))
        self.assertEqual(ConversationState.IDLE, self.bot.state)
        self.assertEqual(self.bot.selected_history, ["testrunner"])
        self.assertEqual(self.bot.active_prompt, None)

    # This is just testing a method in klat_api, not really a chatbot-core test
    # @pytest.mark.timeout(30)
    # def test_10_login_register_new_user(self):
    #     self.bot.logout_klat()
    #     self.assertEqual(self.bot.logged_in, 1)
    #     username = f"testrunner{time.time()}".split(".")[0]
    #     self.bot.username = username
    #     self.bot.password = "testpassword"
    #     self.bot.login_klat(username, "testpassword")
    #     while not self.bot.enable_responses:
    #         time.sleep(1)
    #     self.assertEqual(self.bot.logged_in, 2)
    #     self.assertEqual(self.bot.username, username)

    @pytest.mark.timeout(10)
    def test_11_clean_options(self):
        self.bot.active_prompt = "Test Prompt"
        self.bot.proposed_responses[self.bot.active_prompt] = {self.bot.nick: "This is removed",
                                                               "Other User": "Valid Response",
                                                               "Removed User": "Test Prompt"}
        opts = self.bot._clean_options()
        self.assertIsInstance(opts, dict)
        self.assertNotIn(self.bot.nick, opts.keys())
        self.assertNotIn(self.bot.active_prompt, opts.values())

    @pytest.mark.timeout(10)
    def test_12_valid_add_to_queue(self):
        test_input = ("user", "shout", "cid", "dom", "timestamp")
        self.bot.handle_incoming_shout(test_input[0], test_input[1], test_input[2], test_input[3], test_input[4])
        queued = self.bot.shout_queue.get(timeout=2)
        self.assertEqual(queued, test_input)

    # @pytest.mark.timeout(10)
    # def test_13_add_none_to_queue(self):
    #     self.bot.shout_queue.put(None)
    #     time.sleep(3)
    #     self.assertTrue(self.bot.shout_queue.empty())
    #     self.assertFalse(self.bot.shout_thread.isAlive())

    @pytest.mark.timeout(10)
    def test_14_voting(self):
        self.bot.state = ConversationState.VOTE
        resp = self.bot.vote_response(self.bot.nick)
        self.assertEqual(resp, "abstain")

        resp = self.bot.vote_response("abstain")
        self.assertEqual(resp, "abstain")

        resp = self.bot.vote_response("")
        self.assertIsNone(resp)

        self.assertNotEqual(self.bot.nick, "someuser")
        resp = self.bot.vote_response("someuser")
        self.assertEqual(resp, "someuser")

    @pytest.mark.timeout(10)
    def test_15_histories_length(self):
        self.assertTrue(len(self.bot.request_history) == len(self.bot.participant_history))

    # @pytest.mark.timeout(10)
    # def test_12_shutdown_testing(self):
    #     self.bot.socket.disconnect()
    #     self.assertFalse(self.bot.socket.connected)

    # def test_bots_in_dir(self):
    #     from chatbot_core.utils import get_bots_in_dir
    #     get_bots_in_dir("/home/d_mcknight/PycharmProjects/chatbots/bots/ELIZA")

    @pytest.mark.timeout(30)
    def test_messagebus_connection(self):
        from chatbot_core.utils.bot_utils import init_message_bus
        from threading import Thread
        from ovos_bus_client.client import MessageBusClient

        t, bus = init_message_bus()
        self.assertIsInstance(t, Thread)
        self.assertIsInstance(bus, MessageBusClient)
        self.assertTrue(bus.started_running)

        bus.close()
        t.join(5)

# TODO: Test CLI bot detection, credentials load, etc. DM


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from collections import OrderedDict
from typing import Optional


class FIFOCache:

    def __init__(self, capacity: int = 10):
        """
        Initialize an instance of the first-in-first-out cache with a set capacity
        :param capacity: a maximum number of entries to store in the cache at the same time
        """
        self.cache = OrderedDict()
        self.capacity = capacity

    def get(self, key: str) -> Optional[str]:
        """
        Lookup the cache using a string key
        :param key: a key in the key-value pair
        :return: a value associated with the provided key or None
        """
        if key not in self.cache:
            return None
        else:
            return self.cache[key]

    def put(self, key: str, value: str) -> None:
        """
        Put the key-value into cache if key not in the cache already
        :param key: a key to use in the cache dict
        :param value: a value associated with the key
        :return: None
        """
        if key not in self.cache:
            self.cache[key] = value
            self.cache.move_to_end(key)
            if len(self.cache) > self.capacity:
                self.cache.popitem(last=False)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os

from typing import Optional
from chatbot_core.chatbot_abc import ChatBotABC
from ovos_utils.log import LOG


def get_class() -> Optional[type(ChatBotABC)]:
    """
    Get class matching current CHATBOT_VERSION

    :returns Class instance matching current version if any
    """
    from chatbot_core.v1 import ChatBot as ChatBot_v1
    from chatbot_core.v2 import ChatBot as ChatBot_v2

    version = get_current_version()
    LOG.debug(f"version={version}")
    chatbot_versions = {
        1: ChatBot_v1,
        2: ChatBot_v2
    }

    if version not in chatbot_versions:
        raise InvalidVersionError(f"{version} is not a valid version "
                                  f"({set(chatbot_versions.keys())}")
    return chatbot_versions.get(version)


def get_current_version() -> int:
    """
    Get an int representation of the configured Chatbot version to run
    """
    return 2 if os.environ.get('CHATBOT_VERSION',
                               'v1').lower() in ('v2', '2', 'version2') else 1


class InvalidVersionError(Exception):
    """
    Exception raised when invalid chatbots version is specified
    """



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2024 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
import time

from neon_mq_connector.utils import RepeatingTimer
from neon_mq_connector.utils.rabbit_utils import create_mq_callback
from klat_connector.mq_klat_api import KlatAPIMQ
from pika.exchange_type import ExchangeType

from chatbot_core.utils.enum import ConversationState, BotTypes
from chatbot_core.chatbot_abc import ChatBotABC


class ChatBot(KlatAPIMQ, ChatBotABC):
    """MQ-based chatbot implementation"""

    async_consumers_enabled = True

    def __init__(self, *args, **kwargs):
        config, service_name, vhost, bot_type = self.parse_init(*args, **kwargs)
        mq_config = config.get("MQ") or config
        bot_config = config.get("chatbots", {}).get(service_name)
        KlatAPIMQ.__init__(self, mq_config, service_name, vhost)
        ChatBotABC.__init__(self, service_name, bot_config)
        self.bot_type = bot_type
        self.current_conversations = dict()
        self.on_server = True
        self.default_response_queue = 'shout'
        self.shout_thread = RepeatingTimer(function=self._handle_next_shout,
                                           interval=kwargs.get('shout_thread_interval', 10))
        self.shout_thread.start()

    def parse_init(self, *args, **kwargs) -> tuple:
        """Parses dynamic params input to ChatBot v2"""
        config, service_name, vhost, bot_type = (list(args) + [None] * 4)[:4]
        config: dict = config or kwargs.get('config', {})
        service_name: str = service_name or kwargs.get('service_name', 'undefined_service')
        vhost: str = vhost or kwargs.get('vhost', '/')
        bot_type: repr(BotTypes) = bot_type or kwargs.get('bot_type', BotTypes.SUBMIND)
        return config, service_name, vhost, bot_type

    @create_mq_callback()
    def handle_kick_out(self, body: dict):
        """Handles incoming request to chatbot"""
        cid = body.get('cid', None)
        self.log.info(f'Received kick out from cid: {cid}')
        if cid:
            self.send_announcement(f'{self.nick.split("-")[0]} kicked out', cid)
            self.current_conversations.pop(cid, None)

    @create_mq_callback()
    def handle_invite(self, body: dict):
        """Handles incoming request to chatbot"""
        new_cid = body.pop('cid', None)
        self.log.info(f'Received invitation to cid: {new_cid}')
        if new_cid and not self.current_conversations.get(new_cid, None):
            self.current_conversations[new_cid] = body
            self.set_conversation_state(new_cid, ConversationState.IDLE)
            self.send_announcement(f'{self.nick.split("-")[0]} joined', new_cid)

    def get_conversation_state(self, cid) -> ConversationState:
        return self.current_conversations.get(cid, {}).get('state', ConversationState.IDLE)

    def set_conversation_state(self, cid, state):
        self.log.debug(f'State was: {self.current_conversations.setdefault(cid, {}).get("state", ConversationState.IDLE)}')
        self.current_conversations.setdefault(cid, {})['state'] = state
        self.log.debug(
            f'State become: {self.current_conversations.setdefault(cid, {}).get("state", ConversationState.IDLE)}')

    def _setup_listeners(self):
        super()._setup_listeners()
        self.register_consumer('invitation',
                               self.vhost,
                               f'{self.nick}_invite',
                               self.handle_invite,
                               self.default_error_handler)
        self.register_consumer('kick_out',
                               self.vhost,
                               f'{self.nick}_kick_out',
                               self.handle_kick_out,
                               self.default_error_handler)
        self.register_consumer('incoming_shout',
                               self.vhost,
                               f'{self.nick}_shout',
                               self._on_mentioned_user_message,
                               self.default_error_handler)
        self.register_subscriber('proctor_message',
                                 self.vhost,
                                 self._on_mentioned_user_message,
                                 self.default_error_handler,
                                 exchange='proctor_shout')
        self.register_subscriber('proctor_ping',
                                 self.vhost,
                                 self.handle_proctor_ping,
                                 self.default_error_handler,
                                 exchange='proctor_ping')

    @create_mq_callback()
    def handle_proctor_ping(self, body: dict):
        if body.get('cid') in list(self.current_conversations):
            with self.create_mq_connection(self.vhost) as mq_connection:
                proctor_nick = body.get('nick', '')
                self.log.debug(f'Sending pong to {proctor_nick}')
                self.publish_message(mq_connection,
                                     request_data=dict(nick=self.nick,
                                                       cid=body.get('cid')),
                                     exchange=f'{proctor_nick}_pong',
                                     expiration=3000)
                self.set_conversation_state(body.get('cid'), ConversationState.WAIT)
                self.send_shout(shout='I am ready for the next prompt',
                                cid=body.get('cid'))

    @create_mq_callback()
    def _on_mentioned_user_message(self, body: dict):
        """
            MQ handler for requesting message for current bot
        """
        if body.get('omit_reply'):
            self.log.debug(f"Explicitly requested no response: messageID="
                           f"{body.get('messageID')}")
            return
        if body.get('cid') not in list(self.current_conversations):
            self.log.info(f"Ignoring message "
                          f"(messageID={body.get('messageID')}) outside of "
                          f"current conversations "
                          f"({self.current_conversations})")
            self.log.debug(f"{body}")
            return
        self.handle_incoming_shout(body)

    @create_mq_callback()
    def _on_user_message(self, body: dict):
        """
            MQ handler for requesting message, gets processed in case its addressed to given instance or is a broadcast call
        """
        # Processing message in case its either broadcast or its received is this instance,
        # forbids recursive calls
        if body.get('broadcast', False) or \
                body.get('receiver', None) == self.nick and \
                self.nick != body.get('user', None):
            self._on_mentioned_user_message('', '', '', body)

    def handle_incoming_shout(self, message_data: dict):
        """
            Handles an incoming shout into the current conversation
            :param message_data: data of incoming message
        """
        self.shout_queue.put(message_data)

    @property
    def contextual_api_supported(self) -> bool:
        """ This is a backward compatibility property to ensure gradual migration of V2 subminds API to enable handling of the context """
        # TODO: make it defaulting to True once all the related subminds are migrated (Kirill)
        return False

    def get_chatbot_response(self, cid, message_data, shout, message_sender, is_message_from_proctor,
                             conversation_state) -> dict:
        """
            Makes response based on incoming message data and its context
            :param cid: current conversation id
            :param message_data: message data received
            :param shout: incoming shout data
            :param message_sender: nick of message sender
            :param is_message_from_proctor: is message sender a Proctor
            :param conversation_state: state of the conversation from ConversationStates

            :returns response data as a dictionary, example:
                {
                 "shout": "I vote for Wolfram",
                 "context": {"selected": "wolfram"},
                 "queue": "pat_user_message"
                }
        """
        response = {'shout': '', 'context': {}, 'queue': ''}
        self.log.info(f'Received incoming shout: {shout}')
        if self.contextual_api_supported:
            context_kwargs = {'context': self._build_submind_request_context(message_data=message_data,
                                                                             message_sender=message_sender,
                                                                             is_message_from_proctor=is_message_from_proctor,
                                                                             conversation_state=conversation_state)}
        else:
            context_kwargs = {}
        if not is_message_from_proctor:
            response['shout'] = self.ask_chatbot(user=message_sender,
                                                 shout=shout,
                                                 timestamp=str(message_data.get('timeCreated', int(time.time()))),
                                                 **context_kwargs)
        else:
            response['to_discussion'] = '1'
            response['conversation_state'] = conversation_state
            message_sender = BotTypes.PROCTOR

            self.set_conversation_state(cid, conversation_state)
            if conversation_state == ConversationState.RESP:
                response['shout'] = self.ask_chatbot(user=message_sender,
                                                     shout=shout,
                                                     timestamp=str(message_data.get('timeCreated', int(time.time()))),
                                                     **context_kwargs)
            elif conversation_state == ConversationState.DISC:
                options: dict = message_data.get('proposed_responses', {})
                response['shout'] = self.ask_discusser(options, **context_kwargs)
            elif conversation_state == ConversationState.VOTE:
                selected = self.ask_appraiser(options=message_data.get('proposed_responses', {}), **context_kwargs)
                response['shout'] = self.vote_response(selected)
                if 'abstain' in response['shout'].lower():
                    selected = "abstain"
                response['context']['selected'] = selected
            elif conversation_state == ConversationState.WAIT:
                response['shout'] = 'I am ready for the next prompt'
            response['context']['prompt_id'] = message_data.get('prompt_id', '')
        return response

    @staticmethod
    def _build_submind_request_context(message_data: dict,
                                       message_sender: str,
                                       is_message_from_proctor: bool,
                                       conversation_state: ConversationState) -> dict:
        return {
            'prompt_id': message_data.get('prompt_id', ''),
            'message_sender': message_sender,
            'is_message_from_proctor': is_message_from_proctor,
            'conversation_state': conversation_state,
        }

    def handle_shout(self, message_data: dict, skip_callback: bool = False):
        """
            Handles shout for bot. If receives response - emits message into "bot_response" queue

            :param message_data: dict containing message data received
            :param skip_callback: to skip callback after handling shoult (default to False)
        """
        self.log.info(f'Message data: {message_data}')
        shout = message_data.get('shout') or message_data.get('messageText', '')
        cid = message_data.get('cid', '')
        conversation_state = ConversationState(message_data.get('conversation_state', 0))
        message_sender = message_data.get('nick', 'anonymous')
        is_message_from_proctor = self._user_is_proctor(message_sender)
        if shout:
            response = self.get_chatbot_response(cid=cid, message_data=message_data,
                                                 shout=shout, message_sender=message_sender,
                                                 is_message_from_proctor=is_message_from_proctor,
                                                 conversation_state=conversation_state)
            shout = response.get('shout', None)
            if shout and not skip_callback:
                self.log.info(f'Sending response: {response}')
                prompt_id = response.get('context', {}).get('prompt_id')
                self.send_shout(shout=shout,
                                responded_message=message_data.get('messageID', ''),
                                cid=cid,
                                to_discussion=response.get('to_discussion', '0'),
                                queue_name=response.get('queue', None),
                                context=response.get('context', None),
                                is_announcement=response.get('is_announcement', False),
                                prompt_id=prompt_id,
                                **response.get('kwargs', {}))
            else:
                self.log.warning(
                    f'{self.nick}: No response was sent as no data was received from message data: {message_data}')
        else:
            self.log.warning(f'{self.nick}: Missing "shout" in received message data: {message_data}')

    def _on_connect(self):
        """Emits fanout message to connection exchange once connecting"""
        self.send_shout(shout='hello',
                        context={'version': os.environ.get('SERVICE_VERSION', 'undefined'),
                                 'bot_type': self.bot_type},
                        exchange='connection')

    def _on_disconnect(self):
        """Emits fanout message to connection exchange once disconnecting"""
        self.send_shout(shout='bye',
                        exchange='disconnection')

    def sync(self, vhost: str = None, exchange: str = None, queue: str = None, request_data: dict = None):
        """
            Periodical notification message to be sent into MQ,
            used to notify other network listeners about this service health status

            :param vhost: mq virtual host (defaults to self.vhost)
            :param exchange: mq exchange (defaults to base one)
            :param queue: message queue prefix (defaults to self.service_name)
            :param request_data: data to publish in sync
        """
        curr_time = int(time.time())
        self.log.debug(f'{curr_time} Emitting sync message from {self.nick}')
        self._on_connect()

    def discuss_response(self, shout: str, cid: str = None):
        """
        Called when a bot has some discussion to share
        :param shout: Response to post to conversation
        :param cid: mentioned conversation id
        """
        if self.get_conversation_state(cid) != ConversationState.DISC:
            self.log.warning(f"Late Discussion! {shout}")
        elif not shout:
            self.log.warning(f"Empty discussion provided! ({self.nick})")

    def on_vote(self, prompt_id: str, selected: str, voter: str):
        pass

    def on_discussion(self, user: str, shout: str):
        pass

    def on_proposed_response(self):
        pass

    def on_selection(self, prompt: str, user: str, response: str):
        pass

    def on_ready_for_next(self, user: str):
        pass

    def at_chatbot(self, user: str, shout: str, timestamp: str) -> str:
        pass

    def ask_proctor(self, prompt: str, user: str, cid: str, dom: str):
        pass

    def ask_chatbot(self, user: str, shout: str, timestamp: str, context: dict = None) -> str:
        pass

    def ask_history(self, user: str, shout: str, dom: str, cid: str) -> str:
        pass

    def ask_appraiser(self, options: dict, context: dict = None) -> str:
        pass

    def ask_discusser(self, options: dict, context: dict = None) -> str:
        pass

    def _send_first_prompt(self):
        pass

    def send_shout(self, shout, responded_message=None, cid: str = '', dom: str = '',
                   queue_name='',
                   exchange='',
                   broadcast: bool = True,
                   context: dict = None,
                   prompt_id='',
                   **kwargs) -> str:
        """
            Convenience method to emit shout via MQ with extensive instance properties

            :param shout: response message to emit
            :param responded_message: responded message if any
            :param cid: id of desired conversation
            :param dom: domain name
            :param queue_name: name of the response mq queue
            :param exchange: name of mq exchange
            :param broadcast: to broadcast shout (defaults to True)
            :param context: message context to pass along with response
            :param prompt_id: id of prompt to refer shout to

            :returns generated shout id
        """
        conversation_state = self.get_conversation_state(cid)
        if isinstance(conversation_state, ConversationState):
            conversation_state = conversation_state.value
        queue_name = queue_name or self.default_response_queue
        if broadcast:
            # prohibits fanouts to default exchange for consistency
            exchange = exchange or queue_name
            queue_name = ''
            exchange_type = ExchangeType.fanout.value
        else:
            exchange_type = ExchangeType.direct.value

        kwargs.setdefault('omit_reply', False)
        kwargs.setdefault('no_save', False)

        return self._send_shout(
            queue_name=queue_name,
            exchange=exchange,
            exchange_type=exchange_type,
            message_body={
                'nick': self.nick,
                'bot_type': self.bot_type,
                'service_name': self.service_name,
                'cid': cid,
                'dom': dom,
                'conversation_state': conversation_state,
                'responded_shout': responded_message,
                'shout': shout,
                'context': context or {},
                'prompt_id': prompt_id,
                'time': str(int(time.time())),
                **kwargs})

    def send_announcement(self, shout, cid, **kwargs):
        return self.send_shout(shout=shout,
                               cid=cid,
                               is_announcement='1',
                               **kwargs)

    def vote_response(self, response_user: str, cid: str = None):
        """
            For V2 it is possible to participate in discussions for multiple conversations
            but no more than one discussion per conversation.
        """
        if cid and self.get_conversation_state(cid) != ConversationState.VOTE:
            self.log.warning(f"Late Vote! {response_user}")
            return ''
        elif not response_user:
            self.log.error("Null response user returned!")
            return ''
        elif response_user == "abstain" or response_user in (self.nick, self.service_name):
            # self.self.log.debug(f"Abstaining voter! ({self.nick})")
            return "I abstain from voting"
        else:
            return f"I vote for {response_user}"

    def _handle_next_shout(self):
        """
            Called recursively to handle incoming shouts synchronously
        """
        next_message_data = self.shout_queue.get()
        while next_message_data:
            self.handle_shout(next_message_data)
            next_message_data = self.shout_queue.get()

    def _pause_responses(self, duration: int = 5):
        pass

    def pre_run(self, **kwargs):
        self._setup_listeners()

    def shutdown(self):
        self.shout_thread.cancel()
        self.shout_thread.join()



import os
import time
import unittest

from chatbot_core.utils.bot_utils import run_mq_bot
from .mocks import ChatBotObserverMock


@unittest.skip(reason="Testing requires MQ server configuration")
class TestV2ProctoredConversation(unittest.TestCase):
    # os.environ['CHATBOT_VERSION'] = 'v2'
    # test_cid = 'test_conversation'
    # test_vhost = '/test_chatbots'
    # participating_bots = ("hodor", "wiz", "pard")
    # facilitators = ("proctor",)
    # for bot in participating_bots + facilitators:
    #     run_mq_bot(bot, test_vhost, run_kwargs={"run_sync": True,
    #                                             "run_consumers": True})
    # time.sleep(3)
    # observer = ChatBotObserverMock(config={},  # TODO: Define this config
    #                                service_name='mock_chatbot_observer')
    # observer.run(run_consumers=True, run_sync=False)

    def setUp(self) -> None:
        for bot in self.participating_bots.values():
            for consumer in bot.consumers.values():
                self.assertTrue(consumer.is_alive())

    def test_01_invite_bots(self):
        for bot in self.participating_bots:
            self.observer.invite_bot(bot_name=bot, cid=self.test_cid)
        time.sleep(12)

    def test_11_send_prompt(self):
        self.observer.send_prompt(msg_text='Is the Earth flat?')

    def test_12_send_mock_prompt_results(self):
        current_cid_data = {'userID': 'a5a8963b9f924bec8d69',
                            'messageText': '@proctor @pard @wiz',
                            'messageID': '3628cdf9-53cd-4e35-f607-0aa534a55b4e',
                            'attachments': [],
                            'timeCreated': 1650926390,
                            'message_id': '4554e4b10a1d43e1b667caacaa3567f2',
                            'prompt': {'cid': '6222367739332855a5cab2b6',
                                       'userID': 'a5a8963b9f924bec8d69',
                                       'messageText': 'what is the size of Pacific Ocean?',
                                       'messageID': '7f61c539-bec2-4151-924a-65187cb56db7',
                                       'attachments': [],
                                       'timeCreated': 1650926421,
                                       'message_id': '76be6fdd9641437d8e3435c722ee68b0',
                                       'prompt_id': '8f292a9683de402a9d65438267e77578'},
                            'is_active': True,
                            'available_subminds': ['pard-73ad4e5cae2049c4862a1ef121f6cd4a',
                                                   'wiz-a8fa6d5d1625421b8bd0d602cef0f639'],
                            'participating_subminds': ['pard-73ad4e5cae2049c4862a1ef121f6cd4a',
                                                       'wiz-a8fa6d5d1625421b8bd0d602cef0f639'],
                            'proposed_responses': {'pard-73ad4e5cae2049c4862a1ef121f6cd4a': 'Good conversation makes for good partners, don’t you think?',
                                                   'wiz-a8fa6d5d1625421b8bd0d602cef0f639': 'I do not have information about "the size"'},
                            'submind_opinions': {'pard-73ad4e5cae2049c4862a1ef121f6cd4a': 'Who wants to be friends?',
                                                 'wiz-a8fa6d5d1625421b8bd0d602cef0f639': "pard-73ad4e5cae2049c4862a1ef121f6cd4a's proposal is most factual. Vote for pard-73ad4e5cae2049c4862a1ef121f6cd4a."},
                            'votes': {'pard-73ad4e5cae2049c4862a1ef121f6cd4a': 'abstain', 'wiz-a8fa6d5d1625421b8bd0d602cef0f639': 'pard-73ad4e5cae2049c4862a1ef121f6cd4a'},
                            'votes_per_submind': {'pard-73ad4e5cae2049c4862a1ef121f6cd4a': ['wiz-a8fa6d5d1625421b8bd0d602cef0f639']},
                            'winner': 'pard-73ad4e5cae2049c4862a1ef121f6cd4a'}
        # TODO: implement Scorekeeper testing
        # scorekeeper_bot.send_shout(shout='!MSG:SAVE_PROMPT_RESULTS',
        #                            cid=cid,
        #                            omit_reply=True,
        #                            conversation_context=current_cid_data)



import json
import re
import time
from typing import Optional

import pika
from ovos_utils.log import LOG
from neon_mq_connector import MQConnector
from neon_utils.socket_utils import b64_to_dict


class ChatBotObserverMock(MQConnector):

    def __init__(self, config: Optional[dict], service_name: str):
        super().__init__(config, service_name)
        self.vhost = '/test_chatbots'
        self.register_consumer(name='user_message',
                               vhost=self.vhost,
                               queue='chatbot_user_message',
                               callback=self.handle_user_message,
                               on_error=self.default_error_handler,
                               auto_ack=False)
        self.connected_bots = {}

    def invite_bot(self, bot_name, cid):
        r = re.compile(bot_name)
        matched_bot = [nick for nick in list(self.connected_bots.items()) if r.match(nick)][0]
        with self.create_mq_connection(self.vhost) as mq_connection:
            self.emit_mq_message(connection=mq_connection,
                                 queue=f'{matched_bot}_invite',
                                 exchange='',
                                 request_data={'cid': cid})
            LOG.info(f'{matched_bot} invited to {cid}')
        return matched_bot

    def send_prompt(self, msg_text):
        msg_text = f'!prompt: {msg_text}'
        with self.create_mq_connection(self.vhost) as mq_connection:
            self.emit_mq_message(connection=mq_connection,
                                 queue='chatbot_user_message',
                                 exchange='',
                                 request_data={'bots': ['proctor'],
                                               'messageText': msg_text})
        LOG.info(f'Prompt: {msg_text} emitted successfully')

    def handle_user_message(self,
                            channel: pika.channel.Channel,
                            method: pika.spec.Basic.Return,
                            properties: pika.spec.BasicProperties,
                            body: bytes):
        """
            Handles messages from users

            :param channel: MQ channel object (pika.channel.Channel)
            :param method: MQ return method (pika.spec.Basic.Return)
            :param properties: MQ properties (pika.spec.BasicProperties)
            :param body: request body (bytes)

        """
        if body and isinstance(body, bytes):
            dict_data = b64_to_dict(body)
            LOG.info(f'Received bot requesting data: {dict_data}')
            requested_bots = json.loads(dict_data.pop('bots', []))
            for requested_nick in requested_bots:
                matched_bot = self.invite_bot(bot_name=requested_nick, cid=dict_data.get('cid'))
                time.sleep(2)
                self.emit_message_to_bot(nick=matched_bot, dict_data=dict_data)
        else:
            raise TypeError(f'Invalid body received, expected: bytes string; got: {type(body)}')

    def emit_message_to_bot(self, nick, dict_data):
        """
            Wrapper for emitting message to the bot

            :param nick: target bot nickname
            :param dict_data: dictionary with the data to emit
        """
        prompt_prefix = '!prompt:'
        with self.create_mq_connection(self.vhost) as mq_connection:
            if dict_data['messageText'].lower().startswith(prompt_prefix):
                esc_regex = re.compile(re.escape(prompt_prefix), re.IGNORECASE)
                dict_data['messageText'] = esc_regex.sub('', dict_data['messageText']).strip()
                LOG.info(f'Received prompting message: {dict_data["messageText"]}')
                self.publish_message(connection=mq_connection,
                                     request_data=dict_data,
                                     exchange='prompt')
            else:
                self.emit_mq_message(connection=mq_connection,
                                     queue=f'{nick}_user_message',
                                     exchange='',
                                     request_data=dict_data)

    def handle_chatbot_connection(self,
                                  channel: pika.channel.Channel,
                                  method: pika.spec.Basic.Return,
                                  properties: pika.spec.BasicProperties,
                                  body: bytes):
        """
            Handles chatbot connection request

            :param channel: MQ channel object (pika.channel.Channel)
            :param method: MQ return method (pika.spec.Basic.Return)
            :param properties: MQ properties (pika.spec.BasicProperties)
            :param body: request body (bytes)

        """
        if body and isinstance(body, bytes):
            dict_data = b64_to_dict(body)
            if dict_data.get('bot_type', None):
                new_nick = dict_data.pop('nick', None)
                if new_nick in list(self.connected_bots):
                    LOG.warning(f'Skipping connection attempt from "{new_nick}": its already connected')
                else:
                    self.connected_bots[new_nick] = dict_data
                    LOG.info(f'{new_nick} added to connected bots')
            else:
                LOG.debug('Skip non-bot connection')

    def handle_chatbot_disconnection(self,
                                     channel: pika.channel.Channel,
                                     method: pika.spec.Basic.Return,
                                     properties: pika.spec.BasicProperties,
                                     body: bytes):
        """
            Handles chatbot disconnection request

            :param channel: MQ channel object (pika.channel.Channel)
            :param method: MQ return method (pika.spec.Basic.Return)
            :param properties: MQ properties (pika.spec.BasicProperties)
            :param body: request body (bytes)

        """
        if body and isinstance(body, bytes):
            dict_data = b64_to_dict(body)
            new_nick = dict_data.pop('nick', None)
            if new_nick in list(self.connected_bots):
                self.connected_bots.pop(new_nick, None)
                LOG.warning(f'{new_nick} removed from connected bots')
            else:
                self.connected_bots[new_nick] = dict_data
                LOG.info(f'{new_nick} added to connected bots')



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
import unittest
from unittest.mock import patch


class ChatbotCoreTests(unittest.TestCase):
    def test_ChatBot_class(self):
        from chatbot_core.v1 import ChatBot as V1
        from chatbot_core.v2 import ChatBot as V2

        # Test valid v2 spec
        os.environ["CHATBOT_VERSION"] = "v2"
        from chatbot_core import ChatBot
        self.assertEqual(ChatBot, V2, "Expected v2")
        # sys.modules.pop("chatbot_core")

        # Test valid v1 spec
        os.environ["CHATBOT_VERSION"] = "v1"
        from chatbot_core import ChatBot
        self.assertEqual(ChatBot, V1, "Expected v1")
        # sys.modules.pop("chatbot_core")

        # Test default
        os.environ.pop("CHATBOT_VERSION")
        from chatbot_core import ChatBot
        self.assertEqual(ChatBot, V1, "Expected v1")

        # Test invalid
        os.environ["CHATBOT_VERSION"] = '3'
        from chatbot_core import ChatBot
        self.assertEqual(ChatBot, V1, "Expected v1")

    @patch("chatbot_core.utils.bot_utils.grammar_check")
    @patch("chatbot_core.utils.bot_utils.clean_up_bot")
    @patch("chatbot_core.utils.bot_utils.generate_random_response")
    def test_backwards_compat(self, random_response, clean_up, grammar):
        import chatbot_core

        from chatbot_core.utils.enum import ConversationControls
        self.assertEqual(chatbot_core.ConversationControls,
                         ConversationControls)

        from chatbot_core.utils.enum import ConversationState
        self.assertEqual(chatbot_core.ConversationState, ConversationState)

        random_response.assert_not_called()
        clean_up.assert_not_called()
        grammar.assert_not_called()

        chatbot_core.generate_random_response("one", test=1)
        random_response.assert_called_once_with("one", test=1)

        chatbot_core.clean_up_bot("two", test=2)
        clean_up.assert_called_once_with("two", test=2)

        chatbot_core.grammar_check("three", test=3)
        grammar.assert_called_once_with("three", test=3)


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from copy import copy


def remove_prefix(prefixed_string: str, prefix: str):
    """
    Removes the specified prefix from the string
    :param prefixed_string: raw string to clean
    :param prefix: prefix to remove
    :return: string with prefix removed
    """
    if prefixed_string.startswith(prefix):
        return prefixed_string[len(prefix):].lstrip()
    return prefixed_string


def enumerate_subminds(subminds: list) -> str:
    """
    Enumerates bots in format of type "submind1(,submind2... and submindN)"
    where N is the number of subminds provided

    :param subminds: list of names to format
    :returns formatted string reflecting list of subminds provided
    """
    if len(subminds) == 0:
        return 'No one'
    if len(subminds) == 1:
        return subminds[0]
    subminds_copy = copy(subminds)
    last_submind = subminds_copy.pop()
    and_str = ", and " if len(subminds_copy) > 1 else " and "
    return f"{', '.join(subminds_copy)}{and_str}{last_submind}"



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from enum import IntEnum


class ConversationControls:
    RESP = " asks us to consider:"
    DISC = "Please Discuss"
    VOTE = "Voting on the response to "
    PICK = "Tallying the votes for the responses to "
    NEXT = "I'm ready for the next prompt."
    HIST = "history"
    WAIT = " may respond to the next prompt."


class ConversationState(IntEnum):
    IDLE = 0  # No active prompt
    RESP = 1  # Gathering responses to prompt
    DISC = 2  # Discussing responses
    VOTE = 3  # Voting on responses
    PICK = 4  # Proctor will select response
    WAIT = 5  # Bot is waiting for the proctor to ask them to respond (not participating)


class BotTypes:
    PROCTOR = 'proctor'
    SUBMIND = 'submind'
    OBSERVER = 'observer'


CONVERSATION_STATE_ANNOUNCEMENTS = {
    ConversationState.RESP: 'Accepting responses from subminds ({interval} seconds)',
    ConversationState.DISC: 'Discussing responses from subminds ({interval} seconds)',
    ConversationState.VOTE: 'Voting for candidate responses ({interval} seconds)',
    ConversationState.PICK: 'Selecting a winner among participants'
}



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
import time

from typing import Optional
from ovos_bus_client import Message, MessageBusClient
from ovos_utils.log import LOG

from chatbot_core.utils.enum import BotTypes
from chatbot_core.utils.bot_utils import init_message_bus
from chatbot_core import ChatBot

if os.environ.get('CHATBOT_VERSION', 'v1') == 'v2':
    LOG.warning('Version 2.0 is currently alpha')


class NeonBot(ChatBot):
    """
    Extensible class to handle a chatbot implemented in custom-conversations skill
    """

    def __init__(self, *args, **kwargs):
        self.bot_type = BotTypes.SUBMIND
        self.response = None
        self.response_timeout = 15
        self.bus: Optional[MessageBusClient] = None
        self.bus_config = kwargs.pop('bus_config', {"host": "167.172.112.7",
                                                    "port": 8181,
                                                    "ssl": False,
                                                    "route": "/core"})
        self.script = kwargs.pop('script', None)
        self.script_ended = False
        self.script_started = False
        self._init_bus()
        self._set_bus_listeners()
        super(NeonBot, self).__init__(*args, **kwargs)

        timeout = time.time() + 60
        while not self.script_started and time.time() < timeout:
            time.sleep(1)
        if self.script_started:
            self.log.debug("Neon Bot Started!")
        else:
            self.log.error("Neon Bot Error!")

    def ask_chatbot(self, nick: str, shout: str, timestamp: str):
        """
        Handles an incoming shout into the current conversation
        :param nick: user associated with shout
        :param shout: text shouted by user
        :param timestamp: formatted timestamp of shout
        """
        self.log.debug(f"ask neon: {shout}")
        # shout_time = datetime.datetime.strptime(timestamp, "%I:%M:%S %p")
        # timestamp = round(shout_time.timestamp())
        self.response = None
        self._send_to_neon(shout, timestamp, self.nick)
        # if not self.on_server:
        timeout = time.time() + self.response_timeout
        while not self.response and time.time() < timeout:
            time.sleep(0.5)
        if not self.response:
            self.log.error(f"No response to script input!")
        return self.response or shout

    def on_login(self):
        self.log.debug("NeonBot on_login")
        while not self.bus:
            self.log.error("Bus not configured yet!")
            time.sleep(1)
        while not self.bus.started_running:
            self.log.error("Bus not running yet!")
            time.sleep(1)
        self._send_to_neon("exit", str(round(time.time())), self.nick)
        self.enable_responses = False
        timeout = time.time() + 5
        while not self.script_ended and time.time() < timeout:
            time.sleep(1)
        self._send_to_neon(f"run my {self.script} script", str(round(time.time())), self.nick)

    def _init_bus(self):
        # self.bus = MessageBusClient(self.bus_config["host"], self.bus_config["port"],
        #                             self.bus_config["route"], self.bus_config["ssl"])
        # t = Thread(target=self.bus.run_forever)
        # t.daemon = True
        # t.start()
        t, self.bus = init_message_bus(self.bus_config)
        return t

    def _set_bus_listeners(self):
        self.bus.on("speak", self._handle_speak)

    def _handle_speak(self, message: Message):
        """
        Forwards a Neon response into a shout by the logged in user in their current conversation
        :param message: messagebus message associated with "speak"
        """
        self.log.debug(message.context)
        if message.context.get("client") == self.instance:
            input_to_neon = message.context.get("cc_data", {}).get("raw_utterance")
            if input_to_neon == "exit":
                self.script_ended = True
            elif input_to_neon == f"run my {self.script} script":
                time.sleep(5)  # Matches timeout in cc skill for intro speak signal to be cleared
                self.script_started = True
                self.enable_responses = True
            elif input_to_neon and self.enable_responses:
                # self.log.debug(f'sending shout: {message.data.get("utterance")}')
                # if self.on_server:
                #     self.propose_response(message.data.get("utterance"))
                # else:
                self.response = message.data.get("utterance")

    def _send_to_neon(self, shout: str, timestamp: str, nick: str = None):
        """
        Send input to Neon for skills processing
        :param shout: shout to evaluate
        :param timestamp: timestamp of shout
        :param nick: user associated with shout
        """
        nick = nick or "nobody"
        data = {
            "raw_utterances": [shout],
            "utterances": [shout],
            "lang": "en-US",
            "session": "api",
            "user": nick  # This is the user "hosting" this api connection
        }
        context = {'client_name': 'neon_bot',
                   'source': 'klat',
                   "ident": f"chatbots_{timestamp}",
                   'destination': ["skills"],
                   "mobile": False,
                   "client": "api",
                   "flac_filename": None,
                   "neon_should_respond": True,
                   "nick_profiles": {},
                   "cc_data": {"speak_execute": shout,
                               "audio_file": None,
                               "raw_utterance": shout
                               },
                   "timing": {"received": time.time()}
                   }
        # Emit to Neon for a response
        self.log.debug(data)
        self.bus.emit(Message("recognizer_loop:utterance", data, context))



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import os
import argparse
import click

from os.path import expanduser, relpath
from ovos_utils.log import LOG, log_deprecation


@click.group("chatbots", cls=click.Group,
             no_args_is_help=True, invoke_without_command=True,
             help="Chatbot Core Commands\n\n"
                  "See also: chatbots COMMAND --help")
@click.option("--version", "-v", is_flag=True, required=False,
              help="Print the current version")
def chatbot_core_cli(version: bool = False):
    if version:
        from chatbot_core.version import __version__
        click.echo(f"chatbot-core version {__version__}")


@chatbot_core_cli.command(help="Start an MQ chatbot")
@click.argument("bot_entrypoint")
def start_mq_bot(bot_entrypoint):
    os.environ['CHATBOT_VERSION'] = 'v2'
    from chatbot_core.utils.bot_utils import run_mq_bot
    run_mq_bot(bot_entrypoint)


@chatbot_core_cli.command(help="Start a local single-bot session")
@click.option("--bot-dir", default=None,
              help="Path to legacy chatbots directory")
def debug_bots(bot_dir):
    from chatbot_core.utils.bot_utils import debug_bots
    bot_dir = expanduser(relpath(bot_dir)) if bot_dir else None
    debug_bots(bot_dir)


@chatbot_core_cli.command(help="Start a local CBF-style conversation")
@click.option("--prompter", "-p", help="ID of Chatbot to submit prompts")
def run_local_discussion(prompter):
    from chatbot_core.utils.bot_utils import run_local_discussion
    run_local_discussion(prompter)


# Below are deprecated entrypoints
def cli_start_mq_bot():
    """
    Entrypoint to start an MQ chatbot
    """
    os.environ['CHATBOT_VERSION'] = 'v2'
    from chatbot_core.utils.bot_utils import run_mq_bot
    log_deprecation("Use `chatbots start-mq-bot`", "3.0.0")
    parser = argparse.ArgumentParser(description="Start a chatbot")
    parser.add_argument("--bot", dest="bot_name",
                        help="Chatbot entrypoint name", type=str)
    args = parser.parse_args()
    run_mq_bot(args.bot_name)


def cli_start_bots():
    """
    Entry Point to start bots from a Console Script
    """
    from chatbot_core.utils.bot_utils import SERVER, start_bots
    log_deprecation("This CLI command is deprecated", "3.0.0")

    parser = argparse.ArgumentParser(description="Start some chatbots")
    parser.add_argument("--domain", dest="domain", default="chatbotsforum.org",
                        help="Domain to connect to "
                             "(default: chatbotsforum.org)", type=str)
    parser.add_argument("--dir", dest="bot_dir",
                        help="Path to chatbots (default: ./)", type=str)
    parser.add_argument("--bot", dest="bot_name",
                        help="Optional bot name to run a single bot only",
                        type=str)
    parser.add_argument("--credentials", dest="cred_file",
                        help="Optional path to YAML credentials", type=str)
    parser.add_argument("--username", dest="username",
                        help="Klat username for a single bot", type=str)
    parser.add_argument("--password", dest="password",
                        help="Klat password for a single bot", type=str)
    parser.add_argument("--server", dest="server", default=SERVER,
                        help=f"Klat server (default: {SERVER})", type=str)
    parser.add_argument("--debug", dest="debug", action='store_true',
                        help="Enable more verbose log output")
    parser.add_argument("--bot-names", dest="bot-names",
                        help="comma separated list of bots to run", type=str)
    parser.add_argument("--exclude", dest="exclude",
                        help="comma separated list of bots to exclude "
                             "from running", type=str)
    parser.add_argument("--handle-restart", dest="handle_restart",
                        default=False,
                        help="True to handle server emit to restart bots",
                        type=bool)
    args = parser.parse_args()

    if args.debug:
        LOG.set_level("DEBUG")

    if args.exclude:
        excluded_bots = [name.strip() for name in args.exclude.split(",")]
    else:
        excluded_bots = None
    LOG.debug(args)
    start_bots(args.domain, args.bot_dir, args.username, args.password,
               args.server, args.cred_file, args.bot_name,
               excluded_bots, args.handle_restart)


def cli_stop_bots():
    """
    Stops all start-klat-bot instances
    """
    from time import sleep
    log_deprecation("This CLI command is deprecated", "3.0.0")

    parser = argparse.ArgumentParser(description="Stop some chatbots")
    parser.add_argument("--server", dest="server", default="",
                        help=f"Klat server (default: None)", type=str)
    args = parser.parse_args()
    if args.server:
        server_to_stop = args.server
    else:
        server_to_stop = None
    import psutil

    procs = {p.pid: p.info for p in psutil.process_iter(['name'])}
    for pid, name in procs.items():
        if name.get("name") == "start-klat-bots" and \
                (not server_to_stop or f"--server={server_to_stop}" in
                 psutil.Process(pid).cmdline()):
            LOG.info(f"Terminating {pid}")
            psutil.Process(pid).terminate()
            sleep(1)
            if psutil.pid_exists(pid) and psutil.Process(pid).is_running():
                LOG.error(f"Process {pid} not terminated!!")
                psutil.Process(pid).kill()


def cli_start_prompter():
    """
    Entry Point to start a prompter bot
    """
    log_deprecation("This CLI command is deprecated", "3.0.0")

    from chatbot_core.utils.bot_utils import SERVER, start_bots
    parser = argparse.ArgumentParser(description="Start a prompter chatbot")
    parser.add_argument("--bot", dest="bot_name",
                        help="Optional bot name to run a single bot only", type=str)
    parser.add_argument("--dir", dest="bot_dir",
                        help="Path to chatbots (default: ./)", type=str)
    parser.add_argument("--username", dest="username",
                        help="Klat username for a single bot", type=str)
    parser.add_argument("--password", dest="password",
                        help="Klat password for a single bot", type=str)
    parser.add_argument("--server", dest="server", default=SERVER,
                        help=f"Klat server (default: {SERVER})", type=str)
    parser.add_argument("--debug", dest="debug", action='store_true',
                        help="Enable more verbose log output")
    parser.add_argument("--handle-restart", dest="handle_restart", default=False,
                        help="True to handle server emit to restart bots", type=bool)
    args = parser.parse_args()

    if args.debug:
        LOG.level = "DEBUG"
    LOG.debug(args)
    start_bots("chatbotsforum.org", args.bot_dir, args.username, args.password,
               args.server, None, args.bot_name, None, args.handle_restart,
               True)


def cli_debug_bots():
    """
    Debug bots in the passed directory
    :param bot_dir: directory containing the bot to test
    """
    import sys
    from chatbot_core.utils.bot_utils import debug_bots

    log_deprecation("Use `chatbots debug-bots`", "3.0.0")

    bot_dir = os.getcwd()
    # Try handling passed directory
    if len(sys.argv) > 1:
        arg_dir = os.path.expanduser(sys.argv[1])
        bot_dir = arg_dir if os.path.exists(arg_dir) else bot_dir

    debug_bots(bot_dir)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from abc import abstractmethod
from threading import Event, Thread
from ovos_utils.log import LOG


class ParlaiBot:
    """Class declaring ParlAI-specific methods"""

    def __init__(self, interactive_script, response_timeout=25,
                 done_string: str = '[DONE]',
                 exit_string: str = '[EXIT]'):
        """
            :param interactive_script: a script that creates a world within the ParlAI framework (for reference, see any
                                            ParlaiBot-extended class in the chatbots package, e.g. TuckerBot)
            :param response_timeout: timeout in seconds for ParlAI world to generate a response for a prompt
            :param done_string: string that signals about episode done
            :param exit_string: string that signals about the finish
        """
        import spacy
        self.nlp_engine = spacy.load("en_core_web_sm")

        self.agent_id = 'local_agent'
        self.event = Event()
        self.parlai_thread = Thread(target=interactive_script, args=(self,), daemon=True)
        self.parlai_thread.start()

        self.current_response = ''
        self.current_shout = ''
        self.finished = False

        self._response_timeout = response_timeout
        self.done_string = done_string
        self.exit_string = exit_string

    # Agent-specific methods
    def observe(self, msg):
        """
        Observe the other bot's action result
        """
        if msg['id'] != 'context':
            self.event.set()
            self.current_response = msg["text"]
        LOG.debug(f'[OUT]: {self.current_response}')

    def act(self):
        """
        Make an action to provide the other agent in the task with an input
        """
        reply = self._construct_reply()
        # save the current shout locally and clear the attribute to prevent parley() without incoming shout
        reply_text = self.current_shout
        self.current_shout = ''
        LOG.debug(f'CURRENT SHOUT {reply_text}')
        # check for episode done
        if self.done_string in reply_text:
            raise StopIteration
        # set reply text
        reply['text'] = reply_text
        # check if finished
        if self.exit_string in reply_text:
            self.finished = True
            raise StopIteration
        return reply

    # Compatibility methods
    def getID(self):
        """
        Return agent_id of the bot as an agent for ParlAI
        """
        return self.agent_id

    def epoch_done(self):
        """
        Informs DD that the epoch is done. Using for exiting the process.
        """
        return self.finished

    def reset(self):
        """
        Required for defining by agent, e.g. for clearing local variables on exit
        """
        pass

    # Helper methods
    @staticmethod
    def _capitalize(resp: str) -> str:
        """
        Capitalize each sentence, and all "I"s if a pronoun.
        :param resp: a response to be capitalized
        :return: capitalized string
        """
        cap_marks = (".", "!", "?")
        needs_cap = True  # the first word should be capitalized as well
        cap_resp = []
        for word in resp.split():
            if needs_cap:
                cap_resp.append(word.capitalize())
                needs_cap = False
            elif word in cap_marks or any([word.endswith(mark) for mark in cap_marks]):
                cap_resp.append(word)
                needs_cap = True
            elif word == "i":
                cap_resp.append("I")
                needs_cap = False
            else:
                cap_resp.append(word)
        return " ".join(cap_resp)

    @staticmethod
    def _fix_spacing(resp: str) -> str:
        """Fix spacing, e.g. no spaces before the full period '.', or before and after an apostrophe.
        :param resp: a phrase to fix"""
        fixed_resp = ''
        for i in range(len(resp)):
            try:
                if resp[i] == " " and resp[i + 1] in (".", "?", "!", "'"):
                    continue
                if resp[i] == " " and resp[i - 1] == "'" and resp[i - 2] != "s":
                    continue
                else:
                    fixed_resp = fixed_resp + resp[i]
            except IndexError:
                continue
        return fixed_resp

    # Abstract helper methods
    @abstractmethod
    def _construct_reply(self):
        """
        Construct a reply using parlai.core.message.Message in a concrete class. This method is a hack around
        ParlAI installation, so this MUST always be defined in child classes
        """
        raise NotImplementedError

    @abstractmethod
    def _lookup_cache(self, key):
        """
        Lookup cache for particular prompt:response pair
        """
        pass

    @abstractmethod
    def _update_cache(self, prompt: str, resp: str) -> None:
        """
        Save the current prompt and resp to cache
        :param prompt: incoming prompt
        :param resp: generated response for prompt
        :return:
        """
        pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import random
import time
import inspect

from abc import ABC, abstractmethod
from queue import Queue
from typing import Optional
from ovos_config.config import Configuration

from neon_utils.log_utils import init_log
from ovos_utils.log import LOG


class ChatBotABC(ABC):
    """Abstract class gathering all the chatbot-related methods children should implement"""

    def __init__(self, bot_id: str, config: dict = None):
        """
        Common chatbot initialization
        @param bot_id: ID of this chatbot, used to read configuration
        @param config: Dict configuration for this chatbot
        """
        self._bot_id = bot_id
        self.bot_config = config or Configuration().get("chatbots",
                                                        {}).get(bot_id) or {}
        self.shout_queue = Queue(maxsize=256)
        self.__log = None

    @property
    def log(self):
        if not self.__log:
            # Copy log to support multiple bots in thread with different names
            self.__log = init_log(
                log_name="chatbots").create_logger(self._bot_id)
        name = f"{self._bot_id} - "
        stack = inspect.stack()
        record = stack[2]
        mod = inspect.getmodule(record[0])
        module_name = mod.__name__ if mod else ''
        name += module_name + ':' + record[3] + ':' + str(record[2])
        self.__log.name = name
        return self.__log

    @abstractmethod
    def parse_init(self, *args, **kwargs) -> tuple:
        """Parses dynamic init arguments on the considered instance class initialization"""
        pass

    @abstractmethod
    def on_vote(self, prompt_id: str, selected: str, voter: str):
        """
        Override in any bot to handle counting votes. Proctors use this to select a response.
        :param prompt_id: id of prompt being voted on
        :param selected: bot username voted for
        :param voter: user who voted
        """
        pass

    @abstractmethod
    def on_discussion(self, user: str, shout: str):
        """
        Override in any bot to handle discussion from other subminds. This may inform voting for the current prompt
        :param user: user associated with shout
        :param shout: shout to be considered
        """
        pass

    @abstractmethod
    def on_proposed_response(self):
        """
        Override in Proctor to check when to notify bots to vote
        """
        pass

    @abstractmethod
    def on_selection(self, prompt: str, user: str, response: str):
        """
        Override in any bot to handle a proctor selection of a response
        :param prompt: input prompt being considered
        :param user: user who proposed selected response
        :param response: selected response to prompt
        """
        pass

    @abstractmethod
    def on_ready_for_next(self, user: str):
        """
        Notifies when a bot is finished handling the current prompt and is ready for the next one. This should happen
        shortly after the proctor selects a response.
        :param user: user who is ready for the next prompt
        """
        pass

    @abstractmethod
    def at_chatbot(self, user: str, shout: str, timestamp: str) -> str:
        """
        Override in subminds to handle an incoming shout that is directed at this bot. Defaults to ask_chatbot.
        :param user: user associated with shout
        :param shout: text shouted by user
        :param timestamp: formatted timestamp of shout
        :return: response from chatbot
        """
        return self.ask_chatbot(user, shout, timestamp)

    @abstractmethod
    def ask_proctor(self, prompt: str, user: str, cid: str, dom: str):
        """
        Override in proctor to handle a new prompt to queue
        :param prompt: Cleaned prompt for discussion
        :param user: user associated with prompt
        :param cid: cid prompt is from
        :param dom: dom prompt is from
        """
        pass

    @abstractmethod
    def ask_chatbot(self, user: str, shout: str, timestamp: str) -> str:
        """
        Override in subminds to handle an incoming shout that requires some response. If no response can be determined,
        return the prompt.
        :param user: user associated with shout
        :param shout: text shouted by user
        :param timestamp: formatted timestamp of shout
        :return: response from chatbot
        """
        pass

    @abstractmethod
    def ask_history(self, user: str, shout: str, dom: str, cid: str) -> str:
        """
        Override in scorekeepers to handle an incoming request for the selection history
        :param user: user associated with request
        :param shout: shout requesting history
        :param dom: domain user shout originated from
        :param cid: conversation user shout originated from
        :return: Formatted string response
        """
        pass

    @abstractmethod
    def ask_appraiser(self, options: dict) -> str:
        """
        Override in bot to handle selecting a response to the given prompt. Vote is for the name of the best responder.
        :param options: proposed responses (botname: response)
        :return: user selected from options or "abstain" for no vote
        """
        pass

    @abstractmethod
    def ask_discusser(self, options: dict) -> str:
        """
        Override in bot to handle discussing options for the given prompt. Discussion can be anything.
        :param options: proposed responses (botname: response)
        :return: Discussion response for the current prompt
        """
        pass

    @staticmethod
    def _shout_is_prompt(shout):
        """
        Determines if the passed shout is a new prompt for the proctor.
        :param shout: incoming shout
        :return: true if shout should be considered a prompt
        """
        return shout.lower().startswith("!prompt:")

    @staticmethod
    def _hesitate_before_response(start_time, timeout: int = 5):
        """
            Applies some hesitation time before response

            :param start_time: initial time
            :param timeout: hesitation timeout
        """
        if time.time() - start_time < timeout:
            # Apply some random wait time if we got a response very quickly
            time.sleep(random.randrange(0, 50) / 10)
        else:
            LOG.debug("Skipping artificial wait!")

    @abstractmethod
    def _send_first_prompt(self):
        """
            Sends an initial prompt to the proctor for a prompter bot
        """
        pass

    @abstractmethod
    def handle_shout(self, *args, **kwargs):
        """
            Handles an incoming shout into the current conversation
            :param user: user associated with shout
            :param shout: text shouted by user
            :param cid: cid shout belongs to
            :param dom: domain conversation belongs to
            :param timestamp: formatted timestamp of shout
        """
        pass

    @staticmethod
    def _user_is_proctor(nick):
        """
        Determines if the passed nick is a proctor.
        :param nick: nick to check
        :return: true if nick belongs to a proctor
        """
        return "proctor" in nick.lower()

    @staticmethod
    def _user_is_prompter(nick):
        """
        Determines if the passed nick is a proctor.
        :param nick: nick to check
        :return: true if nick belongs to a proctor
        """
        return "prompter" in nick.lower()

    @abstractmethod
    def vote_response(self, response_user: str, cid: Optional[str] = None):
        """
            Called when a bot appraiser has selected a response
            :param response_user: bot username associated with chosen response
            :param cid: dedicated conversation id (optional)
        """
        pass

    @abstractmethod
    def _handle_next_shout(self):
        """
        Called recursively to handle incoming shouts synchronously
        """
        pass

    @abstractmethod
    def _pause_responses(self, duration: int = 5):
        """
            Pauses generation of bot responses

            :param duration: seconds to pause
        """
        pass



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending
import os
import unittest


class BotUtilsTests(unittest.TestCase):
    def test_start_bots(self):
        from chatbot_core.utils.bot_utils import start_bots
        # Deprecated method

    def test_generate_random_response(self):
        from chatbot_core.utils.bot_utils import generate_random_response
        # Deprecated method

    def test_debug_bots(self):
        from chatbot_core.utils.bot_utils import debug_bots
        # TODO

    def test_clean_up_bot(self):
        from chatbot_core.utils.bot_utils import clean_up_bot
        # TODO

    def test_find_closest_answer(self):
        from chatbot_core.utils.bot_utils import find_closest_answer
        test_sentence = "This is a statement about testing your code."
        options = {'1': "testing is good",
                   '2': "This is a statement about your code",
                   '3': "This is a statement about testing nothing"}

        random_resp = find_closest_answer('random', test_sentence, options)
        self.assertIn(random_resp, options.keys())

        bleu_response = find_closest_answer('bleu_score', test_sentence,
                                            options)
        self.assertEqual(bleu_response, '3', bleu_response)

        damerau_resp = find_closest_answer('damerau_levenshtein_distance',
                                           test_sentence, options)
        self.assertEqual(damerau_resp, '2', damerau_resp)

        self.assertIsNone(find_closest_answer(options=options))
        self.assertIsNone(find_closest_answer(sentence=test_sentence))
        self.assertIsNone(find_closest_answer("", test_sentence, options))
        self.assertIsNone(find_closest_answer("invalid", test_sentence,
                                              options))

    def test_grammar_check(self):
        from chatbot_core.utils.bot_utils import grammar_check
        # TODO

    def test_find_bot_modules(self):
        from chatbot_core.utils.bot_utils import _find_bot_modules
        # TODO

    def test_run_mq_bot(self):
        from chatbot_core.utils.bot_utils import run_mq_bot
        # TODO

    def test_run_sio_bot(self):
        from chatbot_core.utils.bot_utils import run_sio_bot
        # TODO

    def test_run_all_bots(self):
        from chatbot_core.utils.bot_utils import run_all_bots
        # TODO

    def test_run_local_discussion(self):
        from chatbot_core.utils.bot_utils import run_local_discussion
        # TODO


class CacheTests(unittest.TestCase):
    from chatbot_core.utils.cache import FIFOCache
    cache_size = 3
    cache = FIFOCache(capacity=cache_size)

    def test_init(self):
        from collections import OrderedDict
        self.assertIsInstance(self.cache.cache, OrderedDict)
        self.assertEqual(self.cache.capacity, self.cache_size)

    def test_put_get(self):
        num_put = 10
        for i in range(num_put):
            self.cache.put(str(i), str(i))
        self.assertEqual(len(self.cache.cache), self.cache.capacity)
        for i in range(num_put):
            if i < num_put - self.cache_size:
                self.assertIsNone(self.cache.get(str(i)), self.cache.cache)
            else:
                self.assertEqual(self.cache.get(str(i)), str(i),
                                 self.cache.cache)


class TestConversationUtils(unittest.TestCase):
    def test_create_conversation_cycle(self):
        from chatbot_core.utils.conversation_utils import create_conversation_cycle
        from chatbot_core.utils.enum import ConversationState
        convo = create_conversation_cycle()
        self.assertEqual(next(convo), ConversationState.RESP)
        self.assertEqual(next(convo), ConversationState.DISC)
        self.assertEqual(next(convo), ConversationState.VOTE)
        self.assertEqual(next(convo), ConversationState.PICK)
        self.assertEqual(next(convo), ConversationState.IDLE)
        self.assertEqual(next(convo), ConversationState.RESP)


class TestEnum(unittest.TestCase):
    def test_conversation_controls(self):
        from chatbot_core.utils.enum import ConversationControls
        for c in (ConversationControls.RESP, ConversationControls.DISC,
                  ConversationControls.VOTE, ConversationControls.PICK,
                  ConversationControls.NEXT, ConversationControls.HIST,
                  ConversationControls.WAIT):
            self.assertIsInstance(c, str)

    def test_conversation_state(self):
        from chatbot_core.utils.enum import ConversationState
        for state in ConversationState:
            self.assertIsInstance(state.value, int)

    def test_bot_types(self):
        from chatbot_core.utils.enum import BotTypes
        for b in (BotTypes.PROCTOR, BotTypes.SUBMIND, BotTypes.OBSERVER):
            self.assertIsInstance(b, str)

    def test_conversation_state_announcements(self):
        from chatbot_core.utils.enum import ConversationState, \
            CONVERSATION_STATE_ANNOUNCEMENTS
        for state in (ConversationState.RESP, ConversationState.DISC,
                      ConversationState.VOTE, ConversationState.PICK):
            self.assertIsInstance(CONVERSATION_STATE_ANNOUNCEMENTS[state], str)


class LoggerTests(unittest.TestCase):
    def test_make_logger(self):
        from chatbot_core.utils.logger import make_logger
        from ovos_utils.log import LOG

        # Simple named log
        log = make_logger("test")
        self.assertEqual(log, LOG)
        self.assertEqual(LOG.name, "test")

        # Named log with level override
        log = make_logger("test_2", "ERROR")
        self.assertEqual(log, LOG)
        self.assertEqual(LOG.name, "test_2")
        self.assertEqual(LOG.level, "ERROR")


class StringUtilsTests(unittest.TestCase):
    def test_remove_prefix(self):
        from chatbot_core.utils.string_utils import remove_prefix
        test_string = "This is a test..."
        self.assertEqual(remove_prefix(f"@bot {test_string}", "@bot"),
                         test_string)
        self.assertEqual(remove_prefix(f" {test_string}", '@'),
                         f" {test_string}")
        self.assertEqual(remove_prefix(f"{test_string}{test_string}",
                                       test_string), test_string)

    def test_enumerate_subminds(self):
        from chatbot_core.utils.string_utils import enumerate_subminds
        subminds = list()
        self.assertEqual(enumerate_subminds(subminds), "No one")
        subminds.append("user")
        self.assertEqual(enumerate_subminds(subminds), "user")
        subminds.append("user2")
        self.assertEqual(enumerate_subminds(subminds), "user and user2")
        subminds.append("user3")
        self.assertEqual(enumerate_subminds(subminds), "user, user2, and user3")


class VersionUtilsTests(unittest.TestCase):
    def test_get_class(self):
        from chatbot_core.utils.version_utils import get_class
        from chatbot_core.v1 import ChatBot as V1
        from chatbot_core.v2 import ChatBot as V2

        # Explicit valid option
        os.environ["CHATBOT_VERSION"] = "v1"
        self.assertEqual(get_class(), V1)
        os.environ["CHATBOT_VERSION"] = "v2"
        self.assertEqual(get_class(), V2)

        # Default returns v1
        os.environ.pop("CHATBOT_VERSION")
        self.assertEqual(get_class(), V1)

        # Invalid config returns v1
        os.environ["CHATBOT_VERSION"] = "0"
        self.assertEqual(get_class(), V1)

    def test_get_current_version(self):
        from chatbot_core.utils.version_utils import get_current_version
        # Explicit valid option
        os.environ["CHATBOT_VERSION"] = "v1"
        self.assertEqual(get_current_version(), 1)
        os.environ["CHATBOT_VERSION"] = "v2"
        self.assertEqual(get_current_version(), 2)
        os.environ["CHATBOT_VERSION"] = "2"
        self.assertEqual(get_current_version(), 2)

        # Default returns v1
        os.environ.pop("CHATBOT_VERSION")
        self.assertEqual(get_current_version(), 1)

        # Invalid config returns v1
        os.environ["CHATBOT_VERSION"] = "0"
        self.assertEqual(get_current_version(), 1)


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import inspect
import logging
import os
import random
import time
import sys
import yaml

from typing import Optional, Callable, Dict, List
from multiprocessing import Process, Event, synchronize
from threading import Thread, current_thread
from ovos_bus_client import Message, MessageBusClient
from datetime import datetime
from ovos_utils.xdg_utils import xdg_config_home
from ovos_utils.log import LOG, log_deprecation
from neon_utils.net_utils import get_ip_address

from chatbot_core.chatbot_abc import ChatBotABC
from chatbot_core.v2 import ChatBot as ChatBotV2
from chatbot_core.v1 import ChatBot as ChatBotV1


ip = get_ip_address()
if ip == "64.34.186.120":  # Test
    SERVER = "2222.us"
elif ip == "64.225.115.136":  # Cert
    SERVER = "5555.us"
elif ip == "167.172.112.7":  # Prod
    SERVER = "0000.us"
else:
    # Default external connections to production server
    SERVER = "0000.us"

# active_server = None
runner = Event()


def get_ip_address():
    log_deprecation('Import from `neon_utils.net_utils` directly', "2.3.0")
    return get_ip_address()


def _threaded_start_bot(bot, addr: str, port: int, domain: str, user: str,
                        password: str, event: synchronize.Event,
                        is_prompter: bool):
    """
    Helper function for _start_bot
    """
    from klat_connector import start_socket
    # TODO: Deprecate
    if len(inspect.signature(bot).parameters) == 6:
        instance = bot(start_socket(addr, port), domain, user, password, True,
                       is_prompter)
    elif len(inspect.signature(bot).parameters) == 5:
        if is_prompter:
            LOG.error(f"v2 Bot found, prompter functionality will "
                      f"not be enabled! {bot}")
        instance = bot(start_socket(addr, port), domain, user, password, True)
    else:
        LOG.error(f"Bot params unknown: {inspect.signature(bot).parameters}")
        instance = bot(start_socket(addr, port))
    if is_prompter:  # Send intial prompt if this bot is a prompter
        instance.send_shout(instance.initial_prompt)
    event.clear()
    event.wait()

    # Exit when event is set and then clear event to notify calling function
    instance.exit()
    event.clear()


def _start_bot(bot, addr: str, port: int, domain: str, user: str,
               password: str, is_prompter: bool = False)\
        -> (Process, synchronize.Event):
    """
    Creates a thread and starts the passed bot with passed parameters
    :param bot: ChatBot to instantiate
    :param addr: Server address to connect to
    :param port: Server socketIO port
    :param domain: Starting domain
    :param user: Username to login as
    :param password: Password to login with
    :param is_prompter: True if bot is to generate prompts for the Proctor
    :returns: Process bot instance is attached to
    """
    # TODO: Deprecate
    event = Event()
    event.set()
    thread = Process(target=_threaded_start_bot,
                     args=(bot, addr, port, domain, user, password, event,
                           is_prompter))
    thread.daemon = True
    thread.start()
    while event.is_set():
        time.sleep(0.2)
    return thread, event


def get_bots_in_dir(bot_path: str,
                    names_to_consider: str = None) -> dict:
    """
    Gets all ChatBots in the given directory, imports them, and returns a
        dict of their names to modules.
    :param bot_path: absolute file path containing bots
    :param names_to_consider: limit imported instances to certain list
    :return: dict of bot name:ChatBot object
    """
    names_to_consider = names_to_consider or os.environ.get("bot-names")
    import pkgutil
    log_deprecation("This functionality is deprecated. Bots should specify a "
                    "`neon.plugin.chatbot` entrypoint", "3.0.0")
    bots = {}

    try:
        # Make sure we have a path and not a filename
        bot_path = bot_path if os.path.isdir(bot_path) else \
            os.path.dirname(bot_path)
        # Get all bots in the requested directory
        bot_names = [name for _, name, _ in pkgutil.iter_modules([bot_path])]
        # only specified bot names
        if names_to_consider:
            bot_names = list(set(bot_names) & set(names_to_consider.split(',')))
        if bot_names:
            sys.path.append(bot_path)

            for mod in bot_names:
                module = __import__(mod)
                for name, obj in inspect.getmembers(module, inspect.isclass):
                    if name != "ChatBot" and (issubclass(obj, ChatBotABC) or
                                              (mod in name and
                                               isinstance(obj, type))):
                        bots[mod] = obj
            LOG.debug(bots)
    except Exception as e:
        LOG.error(e)
    return bots


def load_credentials_yml(cred_file: str) -> dict:
    """
    Loads a credentials yml file and returns a dictionary of parsed credentials
    per-module.
    :param cred_file: Input yml file containing credentials for bot modules
    :return: dict of bot modules to usernames and passwords
    """
    log_deprecation(f"This functionality is deprecated. Configuration should "
                    f"be specified in `{xdg_config_home()}/neon/chatbots.yaml`",
                    "3.0.0")
    with open(cred_file, 'r') as f:
        credentials_dict = yaml.safe_load(f)
    return credentials_dict


def _start_bot_processes(bots_to_start: dict, username: str, password: str,
                         credentials: dict, server: str, domain: str) -> list:
    # TODO: Deprecate
    processes = []

    # Start Proctor first if in the list of bots to start
    if "Proctor" in bots_to_start.keys():
        bot = bots_to_start.get("Proctor")
        try:
            user = username or credentials.get("Proctor", {}).get("username")
            password = password or credentials.get("Proctor",
                                                   {}).get("password")
            process, event = _start_bot(bot, server, 8888, domain, user,
                                        password, False)
            processes.append(process)
        except Exception as e:
            LOG.error(e)
            LOG.error(bot)

    # Start a socket for each unique bot, bots handle login names
    for name, bot in bots_to_start.items():
        if name != "Proctor":
            LOG.debug(f"Starting: {name}")
            try:
                user = username or credentials.get(name, {}).get("username")
                password = password or credentials.get(name, {}).get("password")
                process, event = _start_bot(bot, server, 8888, domain, user,
                                            password, False)
                processes.append(process)
            except Exception as e:
                LOG.error(name)
                LOG.error(e)
                LOG.error(bot)
    return processes


def start_bots(domain: str = None, bot_dir: str = None, username: str = None,
               password: str = None, server: str = None, cred_file: str = None,
               bot_name: str = None, excluded_bots: list = None,
               handle_restart: bool = False, is_prompter: bool = False):
    """
    Start all the bots in the given bot_dir and connect them to the given domain
    :param domain: Domain to put bots in
    :param bot_dir: Path containing bots to start
    :param username: Username to login with (or bot name if not defined)
    :param password: Password to login with (or None to connect as guest)
    :param server: Klat server url to connect to
    :param cred_file: Path to a credentials yml file
    :param bot_name: Optional name of the bot to start (None for all bots)
    :param excluded_bots: Optional list of bots to exclude from launching
    :param handle_restart: If true, listens for a restart message from the
        server to restart chatbots
    :param is_prompter: If true, bot sends prompts to the Proctor and handles
        responses
    """
    log_deprecation("This method is deprecated. Bots should be loaded by "
                    "entrypoints.", "3.0.0")
    # TODO: Method for loading v1 bots by entrypoints
    # global active_server
    global runner
    domain = domain or "chatbotsforum.org"
    bot_dir = bot_dir or os.getcwd()
    bot_dir = os.path.expanduser(bot_dir)
    server = server or SERVER
    # active_server = server
    LOG.debug(f"Starting bots on server: {server}")
    bots_to_start = get_bots_in_dir(bot_dir)

    # Catch no bots found
    if len(bots_to_start.keys()) == 0:
        LOG.info(f"No bots in: {bot_dir}")
        for d in os.listdir(bot_dir):
            try:
                if str(d) not in ("__pycache__", "tests", "venv", "torchmoji",
                                  "ParlAI") and not d.startswith(".") \
                        and os.path.isdir(os.path.join(bot_dir, d)):
                    LOG.info(f"Found bots dir {d}")
                    bots_to_start = {**get_bots_in_dir(os.path.join(bot_dir,
                                                                    d)),
                                     **bots_to_start}
            except Exception as e:
                LOG.error(e)

    LOG.info(bots_to_start.keys())
    logging.getLogger("klat_connector").setLevel(logging.WARNING)
    logging.getLogger("tensorflow").setLevel(logging.ERROR)
    # proctor = None

    # Load credentials
    if cred_file:
        cred_file = os.path.expanduser(cred_file)
        if not os.path.isfile(cred_file) and \
                os.path.isfile(os.path.join(os.getcwd(), cred_file)):
            cred_file = os.path.join(os.getcwd(), cred_file)
        elif not os.path.isfile(cred_file):
            cred_file = None
    elif os.path.isfile(os.path.join(os.getcwd(), "credentials.yml")):
        cred_file = os.path.join(os.getcwd(), "credentials.yml")

    LOG.debug(f"Found credentials at: {cred_file}")
    if cred_file:
        credentials = load_credentials_yml(cred_file)
    else:
        credentials = {}
    processes = []

    # Check for specified bot to start
    if bot_name:
        LOG.debug(f"Got requested bot:{bot_name}")
        bot = bots_to_start.get(bot_name)
        if bot:
            bots_to_start = {bot_name: bot}
            try:
                user = username or credentials.get(bot_name,
                                                   {}).get("username")
                password = password or credentials.get(bot_name,
                                                       {}).get("password")
                p, _ = _start_bot(bot, server, 8888, domain, user, password,
                                  is_prompter)
                processes.append(p)
                # bot(start_socket(server, 8888), domain, user, password, True)
            except Exception as e:
                LOG.error(e)
        else:
            LOG.error(f"{bot_name} is not a valid bot!")
            return
    # Else start all bots
    else:
        if excluded_bots:
            # Remove any excluded bots
            for name in excluded_bots:
                if name in bots_to_start.keys():
                    bots_to_start.pop(name)

        processes = _start_bot_processes(bots_to_start, username, password,
                                         credentials, server, domain)

    if handle_restart:
        log_deprecation("Messagebus connections to Neon Core will be "
                        "deprecated", "3.0.0")
        LOG.info(f"Setting restart listener for {server}")
        _listen_for_restart_chatbots(server)
    try:
        # runner = Event()
        while True:
            LOG.info(">>>STARTED<<<")
            # Wait for an event that will never come
            runner.clear()
            runner.wait()
            LOG.info(">>>RESTART REQUESTED<<<")
            # LOG.debug(f"procs={len(processes)} | {processes}")
            for p in processes:
                try:
                    LOG.debug(f"Terminating {p.pid}")
                    # processes.remove(p)
                    p.terminate()
                    time.sleep(1)
                    if p.is_alive():
                        LOG.warning(f"Process {p.pid} not terminated! "
                                    f"Killing..")
                        p.kill()
                        time.sleep(1)
                        if p.is_alive():
                            LOG.error(f"Process {p.pid} still alive!")
                except Exception as e:
                    LOG.error(e)
                    p.kill()
            LOG.debug(f"Processes ended")
            processes = _start_bot_processes(bots_to_start, username, password,
                                             credentials, server, domain)
    except KeyboardInterrupt:
        LOG.info("exiting")
        for p in processes:
            p.join()


def debug_bots(bot_dir: str = None):
    """
    Debug bots in the passed directory
    :param bot_dir: directory containing the bot to test
    """
    # TODO: Generalize this to testing different modules? Leave one method for
    #       selecting a bot and then create an options menu for this interactive
    #       testing, along with automated discusser and appraiser testing.
    #       Automated testing could use pre-built response objects, or run n
    #       other bots and handle their outputs offline
    from klat_connector.mach_server import MachKlatServer
    from ovos_config.config import Configuration
    server = MachKlatServer()
    Configuration()['socket_io'] = {"server": "0.0.0.0",
                                    "port": 8888}
    # TODO: Define alternate `ChatBot` base class with no server dependency
    if bot_dir:
        log_deprecation("Bots should be installed so they may be accessed by "
                        "entrypoint. Specifying a local directory will no "
                        "longer be supported", "2.3.1")
        subminds = get_bots_in_dir(bot_dir)
    else:
        subminds = _find_bot_modules()
    # Options to exit the interactive shell
    stop_triggers = ["bye", "see you later", "until next time",
                     "have a great day", "goodbye"]
    running = True
    while running:
        try:
            print(f'BOTS: {subminds.keys()}.\n'
                  f'Please choose a bot to talk to')
            bot_name = input('[In]: ')
            if bot_name in subminds:
                bot = run_sio_bot(bot_name)
                while running:
                    utterance = input('[In]: ')
                    response = bot.ask_chatbot('Tester', utterance,
                                               datetime.now().strftime(
                                                   "%I:%M:%S %p"))
                    print(f'[Out]: {response}')
                    if utterance.lower() in stop_triggers:
                        running = False
                        LOG.debug("STOP RUNNING")
            else:
                print(f'BOTS: {subminds.keys()}.\nThis bot does not exist.'
                      f' Please choose a valid bot to talk to')
        except KeyboardInterrupt:
            running = False
            LOG.warning("STOP RUNNING")
        except EOFError:
            running = False
    server.shutdown_server()
    LOG.info("Done Running")


def clean_up_bot(bot: ChatBotABC):
    """
    Performs any standard cleanup for a bot on destroy
    :param bot: ChatBot instance to clean up
    """
    from chatbot_core.v1 import ChatBot as V1

    if not isinstance(bot, ChatBotABC):
        raise TypeError(f"Expected ChatBot, got: {type(bot)}")
    if isinstance(bot, V1):
        bot.socket.disconnect()
    if hasattr(bot, "shout_queue"):
        bot.shout_queue.put(None)
    if hasattr(bot, "shout_thread"):
        if current_thread() != bot.shout_thread:
            bot.shout_thread.join(0)
    if hasattr(bot, "bus"):
        bot.bus.close()


def _restart_chatbots(message: Message):
    """
    Messagebus handler to restart chatbots on a server
    :param message: Message associated with request
    """
    # TODO: Deprecate in 3.0.0
    global runner
    LOG.debug(f"Restart received: {message.data} | {message.context}")
    runner.set()


def _listen_for_restart_chatbots(server: str):
    """
    Registers a messagebus listener to restart chatbots for the given server
    :param server: base url of the klat server messagebus to listen to
    """
    # TODO: Deprecate in 3.0.0
    if server == "2222.us":
        host = "64.34.186.120"
    elif server == "5555.us":
        host = "64.225.115.136"
    else:
        host = "167.172.112.7"
    LOG.debug(f"Listening for restart message on {host}")
    bus_config = {"host": host,
                  "port": 8181,
                  "ssl": False,
                  "route": "/core"}
    thread, bus = init_message_bus(bus_config)
    bus.on("restart chatbots", _restart_chatbots)


def init_message_bus(bus_config: dict = None) -> (Thread, MessageBusClient):
    """
    Connects to a Neon Core messagebus and returns the thread and bus.
    :param bus_config: messagebus configuration to use
    :return: Thread, messagebus object
    """
    log_deprecation("Messagebus connections to Neon Core will be deprecated",
                    "3.0.0")
    bus_config = bus_config or {"host": "167.172.112.7",
                                "port": 8181,
                                "ssl": False,
                                "route": "/core"}
    bus = MessageBusClient(bus_config["host"], bus_config["port"],
                           bus_config["route"], bus_config["ssl"])
    t = bus.run_in_thread()
    bus.connected_event.wait(10)
    LOG.info(f"Connected to Messagebus at: {bus_config['host']}")
    return t, bus


def generate_random_response(from_iterable: iter):
    """
    Generates some random bot response from the given options
    :param from_iterable: source iterable to get random value from
    """
    log_deprecation("Use `random.choice` directly", "3.0.0")
    return random.choice(from_iterable)


def find_closest_answer(algorithm: str = 'random', sentence: str = None,
                        options: dict = None) -> Optional[str]:
    """
    Determines which option is most similar to an input sentence
    :param algorithm: algorithm considered
    :param sentence: base sentence
    :param options: dict of option ID to response for comparison with `sentence`
    :returns: `option` ID with a value closest to `sentence`, None if requested
              algorithm fails
    """
    if not sentence:
        LOG.warning('Empty sentence supplied')
        return None
    if not options or len(options.keys()) == 0:
        LOG.warning('No options provided')
        return None
    if algorithm == 'random':
        closest_answer = random.choice(list(options.keys()))
    elif algorithm == 'bleu_score':
        try:
            import nltk
            nltk.download('punkt')
            from nltk import word_tokenize
            from nltk.translate.bleu_score import sentence_bleu
        except ImportError:
            LOG.warning("`nltk` not installed. install "
                        "`chatbot-core[lang]` to install NLU packages.")
            return None
        bleu_scores = []
        response_tokenized = word_tokenize(sentence.lower())
        for option in options.keys():
            opinion_tokenized = word_tokenize(options[option].lower())
            if len(opinion_tokenized) > 0:
                if min(len(response_tokenized), len(opinion_tokenized)) < 4:
                    weighting = 1.0 / min(len(response_tokenized),
                                          len(opinion_tokenized))
                    weights = tuple([weighting] * min(len(response_tokenized),
                                                      len(opinion_tokenized)))
                else:
                    weights = (0.25, 0.25, 0.25, 0.25)
                bleu_scores.append((option, sentence_bleu([response_tokenized],
                                                          opinion_tokenized,
                                                          weights=weights)))
        max_score = max([x[1] for x in
                         bleu_scores]) if len(bleu_scores) > 0 else 0
        closest_answer = random.choice(list(filter(lambda x: x[1] == max_score,
                                                   bleu_scores)))[0]
        LOG.info(f'Closest answer is {closest_answer}')
    elif algorithm == 'damerau_levenshtein_distance':
        closest_distance = None
        closest_answer = None
        try:
            import jellyfish
            for option in options.items():
                distance = jellyfish.damerau_levenshtein_distance(option[1],
                                                                  sentence)
                if not closest_distance or closest_distance > distance:
                    closest_distance = distance
                    closest_answer = option[0]
            LOG.info(f'Closest answer is {closest_answer}')
        except ImportError:
            jellyfish = None
            LOG.warning("`jellyfish` not installed. install "
                        "`chatbot-core[lang]` to install NLU packages.")
        except Exception as e:
            LOG.error(e)
    else:
        LOG.error(f'Unknown algorithm supplied:{algorithm}')
        return None
    return closest_answer


def grammar_check(func: Callable):
    """
    Decorator to add spelling/grammar checks to a function's output
    :param func: function to consider
    """
    try:
        from autocorrect import Speller
        spell = Speller()
    except ImportError:
        Speller = None
        LOG.error("autocorrect module not available. Install "
                  "`chatbot-core[extra-lgpl]` to use autocorrect.")
        spell = None

    def wrapper(*args, **kwargs):
        LOG.debug("Entered decorator")
        output = func(*args, **kwargs)
        if output and spell:
            LOG.debug(f"Received output: {output}")
            output = spell(output)
            LOG.debug(f"Processed output: {output}")
        return output

    return wrapper


def _find_bot_modules() -> Dict[str, type(ChatBotABC)]:
    """
    Method for locating all installed chatbots by entrypoint.
    """
    try:
        from importlib_metadata import entry_points
        bot_entrypoints = entry_points(group="neon.plugin.chatbot")
    except ImportError:
        entry_points = None
        import pkg_resources
        bot_entrypoints = pkg_resources.iter_entry_points("neon.plugin.chatbot")

    return {entry.name: entry.load() for entry in bot_entrypoints}


def run_mq_bot(chatbot_name: str, vhost: str = '/chatbots',
               run_kwargs: dict = None, init_kwargs: dict = None) -> ChatBotV2:
    """
    Get an initialized MQ Chatbot instance
    @param chatbot_name: chatbot entrypoint name and configuration key
    @param vhost: MQ vhost to connect to (default /chatbots)
    @param run_kwargs: kwargs to pass to chatbot `run` method
    @param init_kwargs: extra kwargs to pass to chatbot `__init__` method
    @returns: Started ChatBotV2 instance
    """
    from neon_utils.log_utils import init_log
    init_log(log_name=chatbot_name)
    os.environ['CHATBOT_VERSION'] = 'v2'
    run_kwargs = run_kwargs or dict()
    init_kwargs = init_kwargs or dict()
    bots = _find_bot_modules()
    clazz = bots.get(chatbot_name)
    if init_kwargs.get('config'):
        LOG.info(f"Config specified: {init_kwargs['config']}")
    if not clazz:
        raise RuntimeError(f"Requested bot `{chatbot_name}` not found in: "
                           f"{list(bots.keys())}")
    bot = clazz(service_name=chatbot_name, vhost=vhost, **init_kwargs)
    LOG.info(f"Starting {chatbot_name}")
    bot.run(**run_kwargs)
    LOG.info(f"Started {chatbot_name}")
    return bot


def run_sio_bot(chatbot_name: str, domain: str = None,
                is_prompter: bool = False) -> ChatBotV1:
    """
    Get an initialized SIO Chatbot instance
    @param chatbot_name: chatbot entrypoint name and configuration key
    @param domain: Initial domain to enter
    @param is_prompter: If true, submit prompts rather than contribute responses
    @returns: Started ChatBotV2 instance
    """
    from ovos_config.config import Configuration
    from klat_connector import start_socket
    sio_config = Configuration().get("socket_io", {})
    os.environ['CHATBOT_VERSION'] = 'v1'
    domain = domain or "chatbotsforum.org"
    bots = _find_bot_modules()
    clazz = bots.get(chatbot_name)
    if not clazz:
        raise RuntimeError(f"Requested bot `{chatbot_name}` not found in: "
                           f"{list(bots.keys())}")
    sock = start_socket(sio_config.get("server"), sio_config.get("port"))
    bot = clazz(socket=sock, domain=domain, is_prompter=is_prompter)
    LOG.info(f"Started {chatbot_name}")
    return bot


def run_all_bots(domain: str = None) -> List[ChatBotABC]:
    """
    Run all installed chatbots, connecting to the configured server, considering
    the value of the `CHATBOT_VERSION` envvar
    """
    bots = _find_bot_modules()
    from chatbot_core.utils.version_utils import get_current_version
    chatbot_version = get_current_version()
    chatbots = list()
    for bot in bots.keys():
        if chatbot_version == 1:
            chatbots.append(run_sio_bot(bot, domain=domain))
        elif chatbot_version == 2:
            chatbots.append(run_mq_bot(bot))
        else:
            from chatbot_core.utils.version_utils import InvalidVersionError
            raise InvalidVersionError(f"Unable to start chatbot with version: "
                                      f"{chatbot_version}")
    return chatbots


def run_local_discussion(prompter_bot: str):
    """
    Run all installed bots locally with a prompter to submit prompts for
    discussion.
    @param prompter_bot: name/entrypoint of bot to be used as a proctor
    """
    import click
    from ovos_config.config import Configuration
    # Override logging
    Configuration()['log_level'] = "ERROR"

    # Start local server
    from klat_connector import start_socket
    from klat_connector.mach_server import MachKlatServer
    server = MachKlatServer()

    # Load all installed subminds and facilitators
    os.environ['CHATBOT_VERSION'] = 'v1'
    bots = _find_bot_modules()
    chatbots = list()
    for name, clazz in bots.items():
        chatbots.append(clazz(socket=start_socket("0.0.0.0"), domain="local",
                              username=name, password=name))

    prompter_clazz = bots.get(prompter_bot)
    prompter = prompter_clazz(socket=start_socket("0.0.0.0"), domain="private",
                              is_prompter=True, username="Prompter",
                              password=prompter_bot)
    chatbots.append(prompter)
    LOG.info("Local Conversation started")
    # Make conversation output readable
    # TODO: prevent log output going to terminal

    def handle_shout(user, shout, cid, dom, timestamp):
        click.echo(f"{user.rjust(max((len(name) for name in bots)))} : {shout}")

    observer = ChatBotV1(socket=start_socket("0.0.0.0"), domain="local")
    observer.handle_shout = handle_shout

    prompter.send_shout("@proctor hello")
    from ovos_utils import wait_for_exit_signal
    wait_for_exit_signal()
    for bot in chatbots:
        bot.exit()
    server.shutdown_server()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import unittest
from logging import Logger

from unittest.mock import patch
from ovos_utils.log import LOG

from .mocks import MockMQ


class ChatBotV1Tests(unittest.TestCase):
    from klat_connector.mach_server import MachKlatServer
    from klat_connector import start_socket
    server = MachKlatServer()
    socket = start_socket(addr="0.0.0.0")

    @patch("chatbot_core.utils.bot_utils.clean_up_bot")
    def test_init(self, clean_up):
        from chatbot_core.chatbot_abc import ChatBotABC
        from chatbot_core.v1 import ChatBot

        domain = "test_domain"
        bot_kwargs = ChatBot(socket=self.socket, domain=domain,
                             username="test", password="pass", on_server=False,
                             is_prompter=False)
        self.assertIsInstance(bot_kwargs, ChatBotABC)
        self.assertEqual(bot_kwargs.socket, self.socket)
        self.assertEqual(bot_kwargs.start_domain, domain)
        self.assertEqual(bot_kwargs.username, "test")
        self.assertEqual(bot_kwargs._bot_id, "test")
        self.assertEqual(bot_kwargs.password, "pass")
        self.assertFalse(bot_kwargs.on_server)
        self.assertFalse(bot_kwargs.is_prompter)
        self.assertTrue(bot_kwargs.enable_responses)
        from chatbot_core.utils.enum import BotTypes
        self.assertEqual(bot_kwargs.bot_type, BotTypes.SUBMIND)
        self.assertTrue(bot_kwargs.shout_thread.is_alive())

        bot_kwargs.exit()
        self.assertEqual(bot_kwargs.shout_queue.qsize(), 0)
        clean_up.assert_called_once_with(bot_kwargs)

        bot_args = ChatBot(self.socket, domain, "test", "")
        self.assertIsInstance(bot_args, ChatBotABC)
        self.assertEqual(bot_args.socket, self.socket)
        self.assertEqual(bot_args.start_domain, domain)
        self.assertEqual(bot_args.username, "test")
        self.assertEqual(bot_args._bot_id, "test")
        self.assertIsNone(bot_args.password)
        self.assertTrue(bot_args.on_server)
        self.assertFalse(bot_args.is_prompter)
        self.assertTrue(bot_args.enable_responses)
        from chatbot_core.utils.enum import BotTypes
        self.assertEqual(bot_args.bot_type, BotTypes.SUBMIND)
        self.assertTrue(bot_args.shout_thread.is_alive())

        bot_args.exit()
        self.assertEqual(bot_args.shout_queue.qsize(), 0)
        clean_up.assert_called_with(bot_args)

    # TODO


class ChatBotV2Tests(unittest.TestCase):
    @patch("chatbot_core.v2.KlatAPIMQ", new=MockMQ)
    def test_init(self):
        from chatbot_core.v2 import ChatBot
        config = {"test": True,
                  "MQ": {"mq_key": "val"},
                  "chatbots": {"test_bot": {"bot_config": True}}}
        from chatbot_core.utils.enum import BotTypes
        bot_kwargs = ChatBot(config=config, service_name="test_bot",
                             vhost="/test", bot_type=BotTypes.OBSERVER)
        self.assertEqual(bot_kwargs.bot_type, BotTypes.OBSERVER)
        self.assertEqual(bot_kwargs._bot_id, "test_bot")
        self.assertEqual(bot_kwargs.config, config["MQ"])
        self.assertEqual(bot_kwargs.bot_config, config["chatbots"]["test_bot"])
        self.assertEqual(bot_kwargs.vhost, "/test")
        self.assertTrue(bot_kwargs.shout_thread.is_alive())
        bot_kwargs.shutdown()
        self.assertFalse(bot_kwargs.shout_thread.is_alive())

        bot_args = ChatBot(config, "args_bot", "/chat")
        self.assertEqual(bot_args.bot_type, BotTypes.SUBMIND)
        self.assertEqual(bot_args._bot_id, "args_bot")
        self.assertEqual(bot_args.config, config["MQ"])
        self.assertEqual(bot_args.bot_config, dict())
        self.assertEqual(bot_args.vhost, "/chat")
        self.assertTrue(bot_args.shout_thread.is_alive())
        bot_args.shutdown()
        self.assertFalse(bot_args.shout_thread.is_alive())
    # TODO


class ChatBotABCTests(unittest.TestCase):
    def test_base_class(self):
        from queue import Queue
        from chatbot_core.chatbot_abc import ChatBotABC
        from .mocks import TestBot
        bot_id = "test"
        test_config = {"config": True}
        bot = TestBot(bot_id, test_config)
        self.assertIsInstance(bot, ChatBotABC)
        self.assertEqual(bot._bot_id, bot_id)
        self.assertEqual(bot.bot_config, test_config)
        self.assertIsInstance(bot.shout_queue, Queue)
        self.assertIsInstance(bot.log, Logger)
        self.assertTrue(bot.log.name.startswith(bot_id))


class NeonTests(unittest.TestCase):
    from chatbot_core.neon import NeonBot
    # TODO Deprecate class


class ParlaiTests(unittest.TestCase):
    from chatbot_core.parlai import ParlaiBot
    # TODO


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

from ovos_utils.log import log_deprecation


def create_conversation_cycle():
    from chatbot_core.utils.conversation_utils import create_conversation_cycle
    log_deprecation("import from `chatbot_core.utils.conversation_utils`",
                    "2.3.1")
    return create_conversation_cycle()


def find_closest_answer(*args, **kwargs):
    from chatbot_core.utils.bot_utils import find_closest_answer
    log_deprecation("import from `chatbot_core.utils.bot_utils`",
                    "2.3.1")
    return find_closest_answer(*args, **kwargs)


def clean_up_bot(*args, **kwargs):
    from chatbot_core.utils.bot_utils import clean_up_bot
    log_deprecation("import from `chatbot_core.utils.bot_utils`",
                    "2.3.1")
    return clean_up_bot(*args, **kwargs)


def get_bots_in_dir(*args, **kwargs):
    from chatbot_core.utils.bot_utils import get_bots_in_dir
    log_deprecation("import from `chatbot_core.utils.bot_utils`",
                    "2.3.1")
    return get_bots_in_dir(*args, **kwargs)


def start_bots(*args, **kwargs):
    from chatbot_core.utils.bot_utils import start_bots
    log_deprecation("import from `chatbot_core.utils.bot_utils`",
                    "2.3.1")
    return start_bots(*args, **kwargs)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
#
# Copyright 2008-2021 Neongecko.com Inc. | All Rights Reserved
#
# Notice of License - Duplicating this Notice of License near the start of any file containing
# a derivative of this software is a condition of license for this software.
# Friendly Licensing:
# No charge, open source royalty free use of the Neon AI software source and object is offered for
# educational users, noncommercial enthusiasts, Public Benefit Corporations (and LLCs) and
# Social Purpose Corporations (and LLCs). Developers can contact developers@neon.ai
# For commercial licensing, distribution of derivative works or redistribution please contact licenses@neon.ai
# Distributed on an "AS IS” basis without warranties or conditions of any kind, either express or implied.
# Trademarks of Neongecko: Neon AI(TM), Neon Assist (TM), Neon Communicator(TM), Klat(TM)
# Authors: Guy Daniels, Daniel McKnight, Regina Bloomstine, Elon Gasper, Richard Leeds
#
# Specialized conversational reconveyance options from Conversation Processing Intelligence Corp.
# US Patents 2008-2021: US7424516, US20140161250, US20140177813, US8638908, US8068604, US8553852, US10530923, US10530924
# China Patent: CN102017585  -  Europe Patent: EU2156652  -  Patents Pending

import setuptools

from os import path, getenv

BASE_PATH = path.abspath(path.dirname(__file__))


def get_requirements(requirements_filename: str):
    requirements_file = path.join(BASE_PATH, "requirements", requirements_filename)
    with open(requirements_file, 'r', encoding='utf-8') as r:
        requirements = r.readlines()
    requirements = [r.strip() for r in requirements if r.strip() and not r.strip().startswith("#")]

    for i in range(0, len(requirements)):
        r = requirements[i]
        if "@" in r:
            parts = [p.lower() if p.strip().startswith("git+http") else p for p in r.split('@')]
            r = "@".join(parts)
            if getenv("GITHUB_TOKEN"):
                if "github.com" in r:
                    r = r.replace("github.com", f"{getenv('GITHUB_TOKEN')}@github.com")
            requirements[i] = r
    return requirements


with open(path.join(BASE_PATH, "README.md"), "r") as f:
    long_description = f.read()

with open(path.join(BASE_PATH, "chatbot_core", "version.py"), "r", encoding="utf-8") as v:
    for line in v.readlines():
        if line.startswith("__version__"):
            if '"' in line:
                version = line.split('"')[1]
            else:
                version = line.split("'")[1]

setuptools.setup(
    name="neon-chatbot-core",
    version=version,
    author="Neongecko",
    author_email="developers@neon.ai",
    description="Core utilities for Klat chatbots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/neongeckocom/chatbot-core",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6',
    entry_points={'console_scripts': ["chatbots=chatbot_core.cli:chatbot_core_cli",
                                      "start-klat-bots=chatbot_core.cli:cli_start_bots",
                                      "stop-klat-bots=chatbot_core.cli:cli_stop_bots",
                                      "debug-klat-bots=chatbot_core.cli:cli_debug_bots",
                                      "start-klat-prompter=chatbot_core.cli:cli_start_prompter",
                                      "start-mq-bot=chatbot_core.cli:cli_start_mq_bot"]},
    install_requires=get_requirements("requirements.txt"),
    extras_require={"lgpl": get_requirements("extra-lgpl.txt"),
                    "lang": get_requirements("extra-lang.txt")}
)
